/*
 * $Header: /Users/pizlo/Repository/tsf/bdb_file_test.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * bdb_file_test.c -- tests BDB file functionality
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf.h"
#include "tsf_test.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

static tsf_bdb_file_t *file;

static void t_reset() {
    unlink("bdb_file_test.dat");
}

static void t_create(tsf_bool_t truncate) {
    TEST_ASSERT_TSFERROR(
        file=tsf_bdb_file_open("bdb_file_test.dat",tsf_true,truncate));
}

static void t_open(tsf_bool_t update) {
    TEST_ASSERT_TSFERROR(
        file=tsf_bdb_file_open("bdb_file_test.dat",update,tsf_false));
}

static void t_close() {
    TEST_ASSERT_TSFERROR(tsf_bdb_file_sync(file));
    tsf_bdb_file_close(file);
}

static void t_destroy() {
    t_close();
    t_reset();
}

static void test_empty() {
    t_reset();
    t_create(tsf_true);
    t_destroy();
}

static void test_empty_reopen() {
    t_reset();
    t_create(tsf_true);
    t_close();
    t_open(tsf_true);
    t_destroy();
}

static void test_empty_reopen_ro() {
    t_reset();
    t_create(tsf_true);
    t_close();
    t_open(tsf_false);
    t_destroy();
}

static void test_no_void(tsf_bool_t reopen,
                         tsf_bool_t update) {
    tsf_buffer_t *key;
    
    t_reset();
    t_create(tsf_true);
    
    if (reopen) {
        t_close();
        t_open(update);
    }
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));

    TET(tsf_bdb_file_get(file,key), TSF_E_ELE_NOT_FOUND);
    
    tsf_buffer_destroy(key);
    
    t_destroy();
}

static void test_void_to_void(tsf_bool_t reopen,
                              tsf_bool_t update) {
    tsf_buffer_t *key,*val;
    
    t_reset();
    t_create(tsf_true);
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));
    TEST_ASSERT_TSFERROR(
        val=tsf_reflect_make_buffer(tsf_reflect_create_void()));
    
    TEST_ASSERT_TSFERROR(tsf_bdb_file_put(file,key,val));
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    if (reopen) {
        t_close();
        t_open(update);
    }
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));

    TEST_ASSERT_TSFERROR(val=tsf_bdb_file_get(file,key));
    
    TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(val))
                ==TSF_TK_VOID);
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    t_destroy();
}

static void test_void_to_int8(tsf_bool_t reopen,
                              tsf_bool_t update) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    
    t_reset();
    t_create(tsf_true);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(100));
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));
    TEST_ASSERT_TSFERROR(val=tsf_reflect_make_buffer(r));
    
    TEST_ASSERT_TSFERROR(tsf_bdb_file_put(file,key,val));
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    tsf_reflect_destroy(r);
    
    if (reopen) {
        t_close();
        t_open(update);
    }
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));

    TEST_ASSERT_TSFERROR(val=tsf_bdb_file_get(file,key));
    
    TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(val))
                ==TSF_TK_INT8);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_from_buffer(val));
    
    TEST_ASSERT(tsf_type_get_kind_code(tsf_reflect_get_type(val))
                ==TSF_TK_INT8);
    
    TEST_ASSERT(tsf_primitive_reflect_get(r,int8_t)==100);
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    tsf_reflect_destroy(r);
    
    t_destroy();
}

static void test_int8_to_void(int8_t begin,
                              int8_t end,
                              tsf_bool_t reopen,
                              tsf_bool_t update) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    int8_t cur;
    
    t_reset();
    t_create(tsf_true);
    
    for (cur=begin;
         cur<end;
         ++cur) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(
            val=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TEST_ASSERT_TSFERROR(tsf_bdb_file_put(file,key,val));
        
        tsf_buffer_destroy(key);
        tsf_buffer_destroy(val);
        tsf_reflect_destroy(r);
    }
    
    if (reopen) {
        t_close();
        t_open(update);
    }
    
    for (cur=begin;
         cur<end;
         ++cur) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(val=tsf_bdb_file_get(file,key));
        
        TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(val))
                    ==TSF_TK_VOID);
        
        tsf_buffer_destroy(key);
        tsf_buffer_destroy(val);
        tsf_reflect_destroy(r);
    }
    
    t_destroy();
}

static void test_put_n_remove_n(uint32_t n,
                                tsf_bool_t reopen_after_put,
                                tsf_bool_t reopen_after_remove,
                                tsf_bool_t update_after_remove) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    int8_t cur;
    
    t_reset();
    t_create(tsf_true);
    
    for (cur=n;cur-->0;) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint32_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(
            val=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TEST_ASSERT_TSFERROR(tsf_bdb_file_put(file,key,val));
        
        tsf_buffer_destroy(key);
        tsf_buffer_destroy(val);
        tsf_reflect_destroy(r);
    }
    
    TEST_ASSERT(tsf_bdb_file_get_type_code_free_list_size(file)
                == 0);
    
    if (reopen_after_put) {
        t_close();
        t_open(tsf_true);
    }
    
    for (cur=n;cur-->0;) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint32_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(val=tsf_bdb_file_remove(file,key));

        tsf_buffer_destroy(key);
        tsf_buffer_destroy(val);
        tsf_reflect_destroy(r);
    }
    
    TEST_ASSERT(tsf_bdb_file_get_type_code_free_list_size(file)
                == 2);
    
    if (reopen_after_remove) {
        t_close();
        t_open(update_after_remove);
    }
    
    for (cur=n;cur-->0;) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint32_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TET(tsf_bdb_file_get(file,key), TSF_E_ELE_NOT_FOUND);

        tsf_buffer_destroy(key);
        tsf_reflect_destroy(r);
    }
    
    TEST_ASSERT(!tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_VOID)));
    TEST_ASSERT(!tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_UINT32)));
    
    if (reopen_after_remove) {
        TEST_ASSERT(tsf_bdb_file_get_type_code_free_list_size(file)
                    == 0);
    } else {
        TEST_ASSERT(tsf_bdb_file_get_type_code_free_list_size(file)
                    == 2);
    }
    
    t_destroy();
}

static void test_put_two_types_remove_first(tsf_bool_t reopen_after_put,
                                            tsf_bool_t reopen_after_remove,
                                            tsf_bool_t update_after_remove) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    
    t_reset();
    t_create(tsf_true);
    
    /* put the first one */
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_string("first"));
    TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(1));
    TEST_ASSERT_TSFERROR(val=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(tsf_bdb_file_put(file,key,val));
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    /* put the second one */
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_string("second"));
    TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint8_t(1));
    TEST_ASSERT_TSFERROR(val=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(tsf_bdb_file_put(file,key,val));
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    if (reopen_after_put) {
        t_close();
        t_open(tsf_true);
    }
    
    /* remove the first one */
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_string("first"));
    TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(val=tsf_bdb_file_remove(file,key));
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    TEST_ASSERT(!tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT(tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_UINT8)));
    TEST_ASSERT(tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_STRING)));
    
    TEST_ASSERT(tsf_bdb_file_get_type_code_free_list_size(file)
                == 1);
    
    if (reopen_after_remove) {
        t_close();
        t_open(update_after_remove);
    }
    
    TEST_ASSERT(!tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT(tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_UINT8)));
    TEST_ASSERT(tsf_bdb_file_has_type(file,
        tsf_type_create(TSF_TK_STRING)));
    
    TEST_ASSERT(tsf_bdb_file_get_type_code_free_list_size(file)
                == 1);
    
    t_destroy();
}

int main(int c,char **v) {
    if (!tsf_bdb_file_supported()) {
        printf("%s: cannot do test because BDB functionality is not "
               "supported.\n", v[0]);
        return 0;
    }
    
    TEST_INIT(c,v);
    
    DO_TEST(test_empty());
    DO_TEST(test_empty_reopen());
    DO_TEST(test_empty_reopen_ro());
    
    DO_TEST(test_no_void(tsf_false,tsf_false));
    DO_TEST(test_no_void(tsf_true,tsf_false));
    DO_TEST(test_no_void(tsf_true,tsf_true));
    
    DO_TEST(test_void_to_void(tsf_false,tsf_false));
    DO_TEST(test_void_to_void(tsf_true,tsf_false));
    DO_TEST(test_void_to_void(tsf_true,tsf_true));
    
    DO_TEST(test_void_to_int8(tsf_false,tsf_false));
    DO_TEST(test_void_to_int8(tsf_true,tsf_false));
    DO_TEST(test_void_to_int8(tsf_true,tsf_true));
    
    DO_TEST(test_int8_to_void(0,1,tsf_false,tsf_false));
    DO_TEST(test_int8_to_void(0,1,tsf_true,tsf_false));
    DO_TEST(test_int8_to_void(0,1,tsf_true,tsf_true));
    
    DO_TEST(test_int8_to_void(0,2,tsf_false,tsf_false));
    DO_TEST(test_int8_to_void(0,2,tsf_true,tsf_false));
    DO_TEST(test_int8_to_void(0,2,tsf_true,tsf_true));
    
    DO_TEST(test_int8_to_void(0,3,tsf_false,tsf_false));
    DO_TEST(test_int8_to_void(0,3,tsf_true,tsf_false));
    DO_TEST(test_int8_to_void(0,3,tsf_true,tsf_true));
    
    DO_TEST(test_put_n_remove_n(1,
                                tsf_false,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(1,
                                tsf_true,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(1,
                                tsf_true,
                                tsf_true,tsf_false));
    DO_TEST(test_put_n_remove_n(1,
                                tsf_true,
                                tsf_true,tsf_true));
    
    DO_TEST(test_put_n_remove_n(2,
                                tsf_false,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(2,
                                tsf_true,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(2,
                                tsf_true,
                                tsf_true,tsf_false));
    DO_TEST(test_put_n_remove_n(2,
                                tsf_true,
                                tsf_true,tsf_true));
    
    DO_TEST(test_put_n_remove_n(10,
                                tsf_false,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(10,
                                tsf_true,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(10,
                                tsf_true,
                                tsf_true,tsf_false));
    DO_TEST(test_put_n_remove_n(10,
                                tsf_true,
                                tsf_true,tsf_true));
    
    DO_TEST(test_put_two_types_remove_first(tsf_false,
                                            tsf_false,tsf_false));
    DO_TEST(test_put_two_types_remove_first(tsf_true,
                                            tsf_false,tsf_false));
    DO_TEST(test_put_two_types_remove_first(tsf_true,
                                            tsf_true,tsf_false));
    DO_TEST(test_put_two_types_remove_first(tsf_true,
                                            tsf_true,tsf_true));
    
    TEST_DONE();
    
    return 0;
}


