/*
 * $Header: /Users/pizlo/Repository/tsf/flock.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * flock.c -- runs some shell command under a file lock
 * by Filip Pizlo, 2005
 */

#include <sys/file.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>

static void usage() {
    printf("Usage: flock shared|exclusive <lock file> <shell command>\n");
    printf("   or: flock shared|exclusive <lock file>\n");
    printf("Description:\n");
    printf("Acquires either a shared or exclusive lock on the given file and runs the\n");
    printf("given shell command.  If the shell command is not given, this program will\n");
    printf("wait simply pause while holding the lock until it is terminated.\n");
    exit(1);
}

int main(int c,char **v) {
    int fd;

    if (c!=3 && c!=4) {
	usage();
    }

    if (strcmp(v[1],"shared") &&
	strcmp(v[1],"exclusive")) {
	fprintf(stderr,"First argument must be 'shared' or 'exclusive'.\n");
	usage();
    }

    fd=open(v[2],O_RDWR);
    if (fd<0) {
	fprintf(stderr,"Could not open %s: %s\n",
		v[2],strerror(errno));
	return 1;
    }

    if (flock(fd,strcmp(v[1],"shared")?LOCK_EX:LOCK_SH)<0) {
	fprintf(stderr,"Could not acquire lock on %s: %s\n",
		v[2],strerror(errno));
	return 1;
    }
    
    if (c==4) {
	int res=system(v[3]);
	if (res<0) {
	    fprintf(stderr,"Got an error in system(): %s\n",
		    strerror(errno));
	    return 1;
	}
	return res;
    }
    
    pause();
    return 0;
}




