/*
 * $Header: /Users/pizlo/Repository/tsf/generation_verify_test.h,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * generation_verify_test.h -- functions for verifying test data using
 *                             generation
 * by Filip Pizlo, 2004, 2005
 *
 * Note: this is not a normal header file.
 */

typedef void verify_data_t;

static void *verify_from_buffer(tsf_type_t *type,
                                tsf_buffer_t *buf) {
    void *result;
    tsf_parser_t *parser;
    TAT(type=tsf_type_dup(type));
    TAT(tsf_native_type_produce_some_mapping(&type,tsf_true));
    TAT(parser=tsf_parser_create(type,
                                 TSF_EN_DEFAULT,
				 TSF_CD_DEFAULT));
    TAT(result=tsf_parser_parse(parser,buf));
    tsf_parser_destroy(parser);
    return result;
}

static void verify_destroy_data(void *data) {
    tsf_region_free(data);
}

#define DEFINE_PRIMITIVE_VERIFY_FUNCS(type_lc,type_uc)          \
    static void verify_##type_lc(void *data,type_lc value) {    \
        TA(*((type_lc*)data)==value);                           \
    }                                                           \
    typedef int define_primitive_verify_funcs_##type_lc

static void verify_bit(void *data,tsf_bool_t value) {
    TA(*((tsf_bool_t*)data)==value);
}

static void verify_void(void *data) {
}

static void verify_string(void *data,
                          uint32_t size) {
    const unsigned char *string;
    uint32_t i;
    string=*((const char**)data);
    TA(string!=NULL);
    TA(strlen(string)==size);
    for (i=0;i<size;++i) {
        TA(string[i]==(i%255)+1);
    }
}

static tsf_buffer_t *verify_any(void *data) {
    return *((tsf_buffer_t**)data);
}

static void verify_array(tsf_type_t *t,
                         void *data,
                         uint32_t total_elements,
                         uint32_t num_elements,
                         void (*cback)(void *data,
                                       uint32_t i,
                                       void *arg),
                         void *arg) {
    uint32_t i,j;
    TAT(t=tsf_type_dup(t));
    TAT(tsf_native_type_produce_some_mapping(&t,tsf_true));
    switch (tsf_type_get_kind_code(tsf_array_type_get_element_type(t))) {
    case TSF_TK_BIT: {
        tsf_native_bitvector_t *bitvector=data;
        TA(bitvector->num_bits==total_elements);
        for (i=0,j=0;j<total_elements;++j) {
            tsf_bool_t val;
            if (bitvector->bits[j>>3]&(1<<(j&7))) {
                val=tsf_true;
            } else {
                val=tsf_false;
            }
            cback(&val,i,arg);
            i=((i+1)%num_elements);
        }
        break;
    }
    case TSF_TK_VOID: {
        tsf_native_void_array_t *array=data;
        TA(array->len==total_elements);
        break;
    }
    default: {
        tsf_native_array_t *array=data;
        uint32_t ele_size;
        uint8_t *ptr;
        TA(array->len==total_elements);
        ele_size=tsf_native_type_get_size(tsf_array_type_get_element_type(t));
        ptr=array->data;
        i=0;
        for (j=0;j<total_elements;++j) {
            cback(ptr,i,arg);
            ptr+=ele_size;
            i=((i+1)%num_elements);
        }
        break;
    }}
}

static void verify_struct(tsf_type_t *t,
                          void *data,
                          void (*cback)(void *data,
                                        uint32_t i,
                                        void *arg),
                          void *arg) {
    uint32_t i;
    uint8_t *ptr=data;
    TAT(t=tsf_type_dup(t));
    TAT(tsf_native_type_produce_some_mapping(&t,tsf_true));
    for (i=0;
         i<tsf_struct_type_get_num_elements(t);
         ++i) {
        tsf_named_type_t *nt=
            tsf_struct_type_get_element(t,i);
        cback(ptr+nt->offset,
              i,
              arg);
    }
}

static void *verify_choice(tsf_type_t *t,
                           void *data,
                           uint32_t value) {
    uint8_t *ptr=data;
    TAT(t=tsf_type_dup(t));
    TAT(tsf_native_type_produce_some_mapping(&t,tsf_true));
    TA(*((uint32_t*)(ptr+
                     tsf_native_choice_type_get_value_offset(t)))
       ==value);
    if (value!=UINT32_MAX) {
        return ptr+
            tsf_native_choice_type_get_data_offset(t);
    }
    return NULL;
}



