/*
 * $Header: /Users/pizlo/Repository/tsf/gpc_instruction.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * gpc_instruction.c -- instruction definition
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "gpc_internal.h"

#include "gpc_instruction_static_size.gen"
#include "gpc_instruction_size.gen"
#include "gpc_instruction_stack_effects.gen"
#include "gpc_instruction_to_string.gen"

/* some things are still not handled by the specification... */

tsf_bool_t gpc_instruction_for_all_back_branches(gpc_cell_t *inst,
                                                 gpc_target_cback_t cback,
                                                 void *arg) {
    switch (*inst) {
        case GPC_I_ADDCOMPJUMP:
            return cback(inst+2,arg);
        case GPC_I_TWO_ADDCOMPJUMP:
            return cback(inst+3,arg);
        case GPC_I_DECCOMPJUMP:
            return cback(inst+1,arg);
        default:
            break;
    }
    
    return tsf_true;
}

tsf_bool_t gpc_instruction_for_all_forward_branches(gpc_cell_t *inst,
                                                    gpc_target_cback_t cback,
                                                    void *arg) {
    gpc_cell_t i;

    switch (*inst) {
        case GPC_I_COMPFAILJUMP_2ND:
        case GPC_I_COMPFAILJUMP:
        case GPC_I_ZEROJUMP:
            return cback(inst+1,arg);
        case GPC_I_TABLEJUMP_LOCAL:
            for (i=0;i<inst[1];++i) {
                if (!cback(inst+2+i,arg)) {
                    return tsf_false;
                }
            }
            break;
        case GPC_I_TABLEJUMP_FIELD:
            for (i=0;i<inst[2];++i) {
                if (!cback(inst+3+i,arg)) {
                    return tsf_false;
                }
            }
            break;
        case GPC_I_JUMP:
            return cback(inst+1,arg);
        default:
            break;
    }
    
    return tsf_true;
}

tsf_bool_t gpc_instruction_for_all_branches(gpc_cell_t *inst,
                                            gpc_target_cback_t cback,
                                            void *arg) {
    if (!gpc_instruction_for_all_back_branches(inst,cback,arg)) {
        return tsf_false;
    }
    if (!gpc_instruction_for_all_forward_branches(inst,cback,arg)) {
        return tsf_false;
    }
    return tsf_true;
}

tsf_bool_t gpc_instruction_is_return(gpc_cell_t inst) {
    switch (inst) {
        case GPC_I_RETURN_TOP_WITH_REGION:
        case GPC_I_RETURN_BUF_WITH_EMPTY_TYPES:
        case GPC_I_RETURN_BUF_WITH_TYPES_FROM_OUT_MAP:
            return tsf_true;
        default:
            return tsf_false;
    }
}

tsf_bool_t gpc_instruction_is_continuous(gpc_cell_t inst) {
    return !gpc_instruction_is_return(inst) &&
           inst != GPC_I_JUMP;
}


