/*
 * $Header: /Users/pizlo/Repository/tsf/gpc_stack_height.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * gpc_stack_height.c -- implementation of stack height analysis
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "gpc_internal.h"

#define C(exp) do {                         \
    if (!(exp)) {                           \
        goto failure;                       \
    }                                       \
} while (0)

typedef struct {
    gpc_cell_t *cell;
    intptr_t height;
} worklist_item_t;

#include "gpc_worklist.h"

static inline tsf_bool_t pop(worklist_t *worklist,
                             gpc_cell_t **cell,
                             intptr_t *height) {
    worklist_item_t *item=worklist_pop(worklist);
    if (item==NULL) {
        return tsf_false;
    }
    
    *cell=item->cell;
    *height=item->height;
    
    return tsf_true;
}

static inline tsf_bool_t push(worklist_t *worklist,
                              gpc_cell_t *cell,
                              intptr_t height) {
    worklist_item_t *item=worklist_push(worklist);
    if (item==NULL) {
        return tsf_false;
    }
    
    item->cell=cell;
    item->height=height;
    
    return tsf_true;
}

struct branch_action_closure {
    worklist_t *worklist;
    intptr_t height;
};

static tsf_bool_t branch_action(gpc_cell_t *cell,
                                void *arg) {
    struct branch_action_closure *closure=arg;
    return push(closure->worklist,(gpc_cell_t*)*cell,closure->height);
}

tsf_st_table *gpc_intable_get_stack_heights(gpc_intable_t *prog,
                                            intptr_t *max_height) {
    tsf_st_table *ret=tsf_st_init_numtable();
    gpc_cell_t *cur;
    intptr_t height;
    worklist_t worklist;
    
    if (ret==NULL) {
        tsf_set_errno("Could not tsf_st_init_numtable()");
        return NULL;
    }
    
    if (!worklist_init(&worklist)) {
        tsf_st_free_table(ret);
        return NULL;
    }
    
    C(push(&worklist,
           gpc_intable_get_stream(prog),
           gpc_intable_get_num_args(prog)));
    
    if (max_height!=NULL) {
        *max_height=0;
    }
    
    while (pop(&worklist,&cur,&height)) {
        intptr_t got_height;
        
        if (tsf_st_lookup(ret,(char*)cur,(void**)&got_height)) {
            if (got_height != height) {
                tsf_set_error(TSF_E_SEMANTIC,
                              "The stack height is different on two "
                              "different approaches to the same piece "
                              "of code");
                goto failure;
            }
            continue;   /* we've seen it so we don't need to do anything
                         * else. */
        }
        
        if (tsf_st_add_direct(ret,(char*)cur,(void*)height)<0) {
            tsf_set_errno("Could not tsf_st_add_direct()");
            goto failure;
        }
        
        if (max_height!=NULL && height>*max_height) {
            *max_height=height;
        }
        
        height+=gpc_instruction_stack_effects(*cur);
        
        {
            struct branch_action_closure closure;
            closure.worklist=&worklist;
            closure.height=height;
            C(gpc_instruction_for_all_branches(cur,branch_action,&closure));
        }
        
        if (gpc_instruction_is_continuous(*cur)) {
            C(push(&worklist,cur+gpc_instruction_size(cur),height));
        }
    }
    
    worklist_done(&worklist);
    
    return ret;
    
failure:
    tsf_st_free_table(ret);
    worklist_done(&worklist);
    
    return NULL;
}


