/*
 * $Header: /Users/pizlo/Repository/tsf/gpc_worklist.h,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * gpc_worklist.h -- implementation of a worklist.  note that this code
 *                   is a template (polymorphic) and that the template
 *                   parameter is 'worklist_item_t'
 * by Filip Pizlo, 2003, 2004, 2005
 *
 * Due to the polymorphism, this should never be included in a header file,
 * and should be included in a module only once.
 */

#include "gpc_internal.h"
#include "tsf_format.h"

struct worklist {
    worklist_item_t *array;
    worklist_item_t *cur;
    size_t size;
};

typedef struct worklist worklist_t;

static inline tsf_bool_t worklist_init(worklist_t *worklist) {
    worklist->array=NULL;
    worklist->cur=NULL;
    worklist->size=0;
    return tsf_true;
}

static inline void worklist_done(worklist_t *worklist) {
    if (worklist->array!=NULL) {
        free(worklist->array);
    }
}

static inline worklist_item_t *worklist_push(worklist_t *worklist) {
    worklist_item_t *ret;
    
    if (worklist->cur==worklist->array+worklist->size) {
        if (worklist->array==NULL) {
            worklist->array=malloc(sizeof(worklist_item_t));
            if (worklist->array==NULL) {
                tsf_set_errno("Could not malloc worklist array");
                return NULL;
            }
            worklist->cur=worklist->array;
        } else {
            worklist_item_t *new_array=realloc(worklist->array,
                                               sizeof(worklist_item_t)*
                                               (1+worklist->size));
            if (new_array==NULL) {
                tsf_set_errno("Could not realloc worklist array");
                return NULL;
            }
            worklist->cur+=new_array-worklist->array;
            worklist->array=new_array;
        }
        worklist->size++;
    }
    
    ret=worklist->cur;
    worklist->cur++;
    return ret;
}

static inline worklist_item_t *worklist_pop(worklist_t *worklist) {
    if (worklist->cur==worklist->array) {
        return NULL;
    }

    return --worklist->cur;
}


