/*
 * $Header: /Users/pizlo/Repository/tsf/misc_test.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * misc_test.c -- tests stuff that doesn't fit anywhere else.
 * by Filip Pizlo, 2004, 2005
 */

#include "tsf.h"
#include "tsf_test.h"

static void test_space_occupied_by_empty_struct() {
    /* the amount of space occupied by an empty struct should be 0! */
    
    tsf_type_t *empty_struct;
    tsf_genrtr_t *generator;
    tsf_parser_t *parser;
    tsf_buffer_t *buf;
    void *result;
    TAT(empty_struct=tsf_type_create_struct(NULL));
    TAT(tsf_native_type_produce_some_mapping(&empty_struct,tsf_true));
    TAT(generator=tsf_generator_create(empty_struct,
				       TSF_EN_DEFAULT,
				       TSF_CD_DEFAULT));
    TAT(parser=tsf_parser_create(empty_struct,
				 TSF_EN_DEFAULT,
				 TSF_CD_DEFAULT));
    TAT(buf=tsf_generator_generate(generator,NULL));
    TAT(result=tsf_parser_parse(parser,buf));
    TA(tsf_region_size(result)==0); /* take that, I_ALLOC_ROOT! */
    tsf_region_free(result);
    tsf_buffer_destroy(buf);
    tsf_parser_destroy(parser);
    tsf_generator_destroy(generator);
    tsf_type_destroy(empty_struct);
}

static void test_type_create_struct_empty() {
    tsf_type_t *type_one,
               *type_two;
    
    TEST_ASSERT_TSFERROR(type_one=tsf_type_create(TSF_TK_STRUCT));
    
    TEST_ASSERT_TSFERROR(type_two=tsf_type_create_struct(NULL));
    
    TEST_ASSERT(tsf_type_compare(type_one, type_two));
}

static void test_type_create_struct_one() {
    tsf_type_t *type_one,
               *type_two;
    
    TEST_ASSERT_TSFERROR(type_one=tsf_type_create(TSF_TK_STRUCT));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "blah",
                               tsf_type_create(TSF_TK_FLOAT)));
    
    TEST_ASSERT_TSFERROR(
        type_two=tsf_type_create_struct("blah",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL));
    
    TEST_ASSERT(tsf_type_compare(type_one, type_two));
}

static void test_type_create_struct_three() {
    tsf_type_t *type_one,
               *type_two;
    
    TEST_ASSERT_TSFERROR(type_one=tsf_type_create(TSF_TK_STRUCT));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "a",
                               tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "b",
                               tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "c",
                               tsf_type_create(TSF_TK_FLOAT)));
    
    TEST_ASSERT_TSFERROR(
        type_two=tsf_type_create_struct("a",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        "b",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        "c",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL));
    
    TEST_ASSERT(tsf_type_compare(type_one, type_two));
}

int main(int c,char **v) {
    TEST_INIT(c,v);
    
    DO_TEST(test_space_occupied_by_empty_struct());
    DO_TEST(test_type_create_struct_empty());
    DO_TEST(test_type_create_struct_one());
    DO_TEST(test_type_create_struct_three());
    
    TEST_DONE();
    
    return 0;
}


