/*
 * $Header: /Users/pizlo/Repository/tsf/pipe_test_post.h,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * pipe_test_post.h -- everything you need at the bottom of your
 *                     pipe test impl.
 * by Filip Pizlo, 2003, 2004, 2005
 *
 * This is a hack.  Don't mistake it for a normal header file.
 */

static tsf_serial_in_man_t *in_man;
static tsf_serial_out_man_t *out_man;
static int my_pipe[2];
static void *reader_arg,*writer_arg;

static void init_pipe(void) {
    tsf_bool_t in_safe_code;
    char template_name[64];
    strcpy(template_name,TSF_TMP_DIRECTORY "/tsf_pipe_testXXXXXXX");
    in_safe_code=tsf_set_in_safe_code(tsf_false);
    TAE((my_pipe[1]=mkstemp(template_name))>=0);
    tsf_set_in_safe_code(in_safe_code);
    TAE((my_pipe[0]=open(template_name,O_RDONLY))>=0);
    unlink(template_name);
    if (reader_to_use==tsf_fd_reader) {
        reader_arg=my_pipe;
    } else if (reader_to_use==tsf_buf_reader_read) {
        TAT(reader_arg=tsf_buf_reader_create(my_pipe[0],
                                             reader_buf_size,
                                             tsf_false));
    } else {
        TF("reader_to_use has invalid value");
    }
    if (writer_to_use==tsf_fd_writer) {
        writer_arg=my_pipe+1;
    } else if (writer_to_use==tsf_buf_writer_write) {
        TAT(writer_arg=tsf_buf_writer_create(my_pipe[1],
                                             writer_buf_size,
                                             tsf_false));
    } else {
        TF("writer_to_use has invalid value");
    }
    TAT(in_man=tsf_serial_in_man_create(reader_to_use,
                                        reader_arg,
                                        NULL));
    TAT(out_man=tsf_serial_out_man_create(writer_to_use,
                                          writer_arg));
}

static verify_data_t *write_and_read(tsf_type_t *t,
                                     create_data_t *data1) {
    verify_data_t *data2;
    tsf_buffer_t *buf;
    TAT(t); /* convenience */
    buf=create_make_buffer(t,data1);
    create_destroy_data(data1);
    TAT(tsf_serial_out_man_write(out_man,
                                 buf));
    tsf_buffer_destroy(buf);
    if (writer_to_use==tsf_buf_writer_write) {
        TAT(tsf_buf_writer_flush(writer_arg));
    }
    TAT(buf=tsf_serial_in_man_read(in_man));
    TET(tsf_serial_in_man_read(in_man),
        TSF_E_EOF);
    TA(tsf_type_compare(t,tsf_buffer_get_type(buf)));
    data2=verify_from_buffer(t,buf);
    tsf_buffer_destroy(buf);
    return data2;
}

static uint32_t cur_inner_elements; /* evil, evil, evil */

#define SIMPLE_ARRAY_TEST(create_func,verify_func) do {             \
        create_data_t *ele[3];                                      \
        tsf_type_t *arr_t;                                          \
        init_pipe();                                                \
        TAT(arr_t=tsf_type_create_array(t));                        \
        values[0]=value1;                                           \
        values[1]=value2;                                           \
        values[2]=value3;                                           \
        ele[0]=create_func(value1);                                 \
        ele[1]=create_func(value2);                                 \
        ele[2]=create_func(value3);                                 \
        verify_array(arr_t,                                         \
                     write_and_read(arr_t,                          \
                                    create_array(t,                 \
                                                 total_elements,    \
                                                 3,                 \
                                                 ele)),             \
                     total_elements,                                \
                     3,                                             \
                     verify_func,                                   \
                     values);                                       \
    } while (0)

#define SIMPLE_STRUCT_TEST(create_func,verify_func) do {            \
        tsf_type_t *struct_type=tsf_type_create_struct(name1,       \
                                                       t,           \
                                                       name2,       \
                                                       t,           \
                                                       name3,       \
                                                       t,           \
                                                       NULL);       \
        create_data_t *ele[3];                                      \
        init_pipe();                                                \
        TAT(struct_type);                                           \
        values[0]=value1;                                           \
        values[1]=value2;                                           \
        values[2]=value3;                                           \
        ele[0]=create_func(value1);                                 \
        ele[1]=create_func(value2);                                 \
        ele[2]=create_func(value3);                                 \
        verify_struct(struct_type,                                  \
                      write_and_read(struct_type,                   \
                                     create_struct(struct_type,     \
                                                   ele)),           \
                      verify_func,                                  \
                      values);                                      \
    } while (0)

#define SIMPLE_CHOICE_TEST(create_func,verify_func) do {        \
        tsf_type_t *choice_t;                                   \
        create_data_t *cr=NULL;                                 \
        verify_data_t *vr;                                      \
        TAT(choice_t=tsf_type_create_choice(name1,t,            \
                                            name2,t,            \
                                            name3,t,            \
                                            NULL));             \
        init_pipe();                                            \
        if (value!=UINT32_MAX) {                                \
            cr=create_func(data);                               \
        }                                                       \
        vr=verify_choice(choice_t,                              \
                         write_and_read(choice_t,               \
                                        create_choice(choice_t, \
                                                      value,    \
                                                      cr)),     \
                         value);                                \
        if (value!=UINT32_MAX) {                                \
            verify_func(vr,data);                               \
        }                                                       \
    } while (0)

#define SIMPLE_ARRAY_WITH_ARRAY_TEST(create_func,verify_func,arg_t) \
    do {                                                            \
        tsf_type_t *arr_t=tsf_type_create_array(t);                 \
        tsf_type_t *arr_arr_t;                                      \
        create_data_t *ele[3];                                      \
        va_list lst;                                                \
        uint32_t i,j;                                               \
        init_pipe();                                                \
        TAT(arr_t);                                                 \
        TAT(arr_arr_t=tsf_type_create_array(arr_t));                \
        cur_inner_elements=inner_elements;                          \
        va_start(lst,inner_elements);                               \
        for (i=0;i<3;++i) {                                         \
            create_data_t *sub_ele[3];                              \
            for (j=0;j<3;++j) {                                     \
                values[i][j]=va_arg(lst,arg_t);                     \
                sub_ele[j]=create_func(values[i][j]);               \
            }                                                       \
            ele[i]=create_array(t,                                  \
                                inner_elements,                     \
                                3,                                  \
                                sub_ele);                           \
        }                                                           \
        verify_array(arr_arr_t,                                     \
                     write_and_read(arr_arr_t,                      \
                                    create_array(arr_t,             \
                                                 outer_elements,    \
                                                 3,                 \
                                                 ele)),             \
                     outer_elements,                                \
                     3,                                             \
                     verify_func,                                   \
                     values);                                       \
    } while (0)

#define DEFINE_PRIMITIVE_TEST(type_lc,type_uc,arg_t)                \
    DEFINE_PRIMITIVE_CREATE_FUNCS(type_lc,type_uc);                 \
    DEFINE_PRIMITIVE_VERIFY_FUNCS(type_lc,type_uc);                 \
    static void test_##type_lc(type_lc value) {                     \
        init_pipe();                                                \
        verify_##type_lc(write_and_read(tsf_type_create(type_uc),   \
                                        create_##type_lc(value)),   \
                         value);                                    \
    }                                                               \
    static void test_any_with_##type_lc(type_lc value) {            \
        tsf_type_t *t=tsf_type_create(type_uc);                     \
        tsf_buffer_t *buf;                                          \
        init_pipe();                                                \
        TAT(t);                                                     \
        buf=create_make_buffer(t,create_##type_lc(value));          \
        buf=verify_any(write_and_read(tsf_type_create(TSF_TK_ANY),  \
                                      create_any(buf)));            \
        verify_##type_lc(verify_from_buffer(t,buf),                 \
                         value);                                    \
    }                                                               \
    static void verify_array_with_##type_lc(verify_data_t *r,       \
                                            uint32_t i,             \
                                            void *arg) {            \
        type_lc *values=arg;                                        \
        verify_##type_lc(r,values[i]);                              \
    }                                                               \
    static void test_array_with_##type_lc(uint32_t total_elements,  \
                                          type_lc value1,           \
                                          type_lc value2,           \
                                          type_lc value3) {         \
        tsf_type_t *t=tsf_type_create(type_uc);                     \
        type_lc values[3];                                          \
        TAT(t);                                                     \
        SIMPLE_ARRAY_TEST(create_##type_lc,                         \
                          verify_array_with_##type_lc);             \
    }                                                               \
    static void                                                     \
    verify_array_with_array_with_##type_lc(verify_data_t *r,        \
                                           uint32_t i,              \
                                           void *arg) {             \
        type_lc (*values)[3]=arg;                                   \
        verify_array(tsf_type_create_array                          \
                     (tsf_type_create(type_uc)),                    \
                     r,                                             \
                     cur_inner_elements,                            \
                     3,                                             \
                     verify_array_with_##type_lc,                   \
                     values[i]);                                    \
    }                                                               \
    static void                                                     \
    test_array_with_array_with_##type_lc(uint32_t outer_elements,   \
                                         uint32_t inner_elements,   \
                                         ...) {                     \
        tsf_type_t *t=tsf_type_create(type_uc);                     \
        type_lc values[3][3];                                       \
        TAT(t);                                                     \
        SIMPLE_ARRAY_WITH_ARRAY_TEST                                \
            (create_##type_lc,                                      \
             verify_array_with_array_with_##type_lc,                \
             arg_t);                                                \
    }                                                               \
    static void verify_struct_with_##type_lc(verify_data_t *r,      \
                                             uint32_t i,            \
                                             void *arg) {           \
        verify_##type_lc(r,                                         \
                         ((type_lc*)arg)[i]);                       \
    }                                                               \
    static void test_struct_with_##type_lc(const char *name1,       \
                                           type_lc value1,          \
                                           const char *name2,       \
                                           type_lc value2,          \
                                           const char *name3,       \
                                           type_lc value3) {        \
        tsf_type_t *t=tsf_type_create(type_uc);                     \
        type_lc values[3];                                          \
        TAT(t);                                                     \
        SIMPLE_STRUCT_TEST(create_##type_lc,                        \
                           verify_struct_with_##type_lc);           \
    }                                                               \
    static void test_choice_with_##type_lc(const char *name1,       \
                                           const char *name2,       \
                                           const char *name3,       \
                                           uint32_t value,          \
                                           type_lc data) {          \
        tsf_type_t *t;                                              \
        TAT(t=tsf_type_create(type_uc));                            \
        SIMPLE_CHOICE_TEST(create_##type_lc,                        \
                           verify_##type_lc);                       \
    }                                                               \
    typedef void define_primitive_test_##type_lc##_##type_uc

DEFINE_PRIMITIVE_TEST(int8_t,TSF_TK_INT8,int);
DEFINE_PRIMITIVE_TEST(uint8_t,TSF_TK_UINT8,unsigned);
DEFINE_PRIMITIVE_TEST(int16_t,TSF_TK_INT16,int);
DEFINE_PRIMITIVE_TEST(uint16_t,TSF_TK_UINT16,unsigned);
DEFINE_PRIMITIVE_TEST(int32_t,TSF_TK_INT32,int32_t);
DEFINE_PRIMITIVE_TEST(uint32_t,TSF_TK_UINT32,uint32_t);
DEFINE_PRIMITIVE_TEST(int64_t,TSF_TK_INT64,int64_t);
DEFINE_PRIMITIVE_TEST(uint64_t,TSF_TK_UINT64,uint64_t);
DEFINE_PRIMITIVE_TEST(float,TSF_TK_FLOAT,double);
DEFINE_PRIMITIVE_TEST(double,TSF_TK_DOUBLE,double);

static void test_bit(tsf_bool_t value) {
    init_pipe();
    verify_bit(write_and_read(tsf_type_create(TSF_TK_BIT),
                              create_bit(value)),
               value);
}

static void test_void(void) {
    init_pipe();
    verify_void(write_and_read(tsf_type_create(TSF_TK_VOID),
                               create_void()));
}

static void test_string(uint32_t size) {
    init_pipe();
    verify_string(write_and_read(tsf_type_create(TSF_TK_STRING),
                                 create_string(size)),
                  size);
}

static void test_any_with_void(void) {
    tsf_type_t *void_type=tsf_type_create(TSF_TK_VOID);
    tsf_buffer_t *buf;
    init_pipe();
    TAT(void_type);
    buf=create_make_buffer(void_type,create_void());
    buf=verify_any(write_and_read(tsf_type_create(TSF_TK_ANY),
                                  create_any(buf)));
    verify_void(verify_from_buffer(void_type,buf));
}

static void test_any_with_bit(tsf_bool_t value) {
    tsf_type_t *bit_type=tsf_type_create(TSF_TK_BIT);
    tsf_buffer_t *buf;
    init_pipe();
    TAT(bit_type);
    buf=create_make_buffer(bit_type,create_bit(value));
    buf=verify_any(write_and_read(tsf_type_create(TSF_TK_ANY),
                                  create_any(buf)));
    verify_bit(verify_from_buffer(bit_type,buf),
               value);
}

static void test_any_with_string(uint32_t size) {
    tsf_type_t *str_type=tsf_type_create(TSF_TK_STRING);
    tsf_buffer_t *buf;
    init_pipe();
    TAT(str_type);
    buf=create_make_buffer(str_type,create_string(size));
    buf=verify_any(write_and_read(tsf_type_create(TSF_TK_ANY),
                                  create_any(buf)));
    verify_string(verify_from_buffer(str_type,buf),
                  size);
}

static void verify_array_with_void(verify_data_t *r,
                                   uint32_t i,
                                   void *arg) {
    verify_void(r);
}

static void test_array_with_void(uint32_t total_elements) {
    tsf_type_t *void_type=tsf_type_create(TSF_TK_VOID);
    tsf_type_t *arr_t;
    create_data_t *ele;
    init_pipe();
    TAT(void_type);
    TAT(arr_t=tsf_type_create_array(void_type));
    ele=create_void();
    verify_array(arr_t,
                 write_and_read(arr_t,
                                create_array(void_type,
                                             total_elements,
                                             1,
                                             &ele)),
                 total_elements,
                 1,
                 verify_array_with_void,
                 NULL);
}

static void verify_array_with_array_with_void(verify_data_t *r,
                                              uint32_t i,
                                              void *arg) {
    tsf_type_t *t=tsf_type_create(TSF_TK_VOID);
    tsf_type_t *arr_t;
    TAT(t);
    TAT(arr_t=tsf_type_create_array(t));
    verify_array(arr_t,
                 r,
                 cur_inner_elements,
                 3,
                 verify_array_with_void,
                 NULL);
}

static void test_array_with_array_with_void(uint32_t outer_elements,
                                            uint32_t inner_elements) {
    tsf_type_t *t=tsf_type_create(TSF_TK_VOID);
    tsf_type_t *arr_t,*arr_arr_t;
    create_data_t *ele[3];
    create_data_t *sub_ele;
    uint32_t i;
    init_pipe();
    TAT(t);
    TAT(arr_t=tsf_type_create_array(t));
    TAT(arr_arr_t=tsf_type_create_array(arr_t));
    cur_inner_elements=inner_elements;
    sub_ele=create_void();
    for (i=0;i<3;++i) {
        ele[i]=create_array(t,
                            inner_elements,
                            1,
                            &sub_ele);
    }
    verify_array(arr_arr_t,
                 write_and_read(arr_arr_t,
                                create_array(arr_t,
                                             outer_elements,
                                             3,
                                             ele)),
                 outer_elements,
                 3,
                 verify_array_with_array_with_void,
                 NULL);
}

static void verify_array_with_bit(verify_data_t *r,
                                  uint32_t i,
                                  void *arg) {
    tsf_bool_t *values=arg;
    verify_bit(r,values[i]);
}
               
static void test_array_with_bit(uint32_t total_elements,
                                tsf_bool_t value1,
                                tsf_bool_t value2,
                                tsf_bool_t value3) {
    tsf_type_t *t=tsf_type_create(TSF_TK_BIT);
    tsf_bool_t values[3];
    TAT(t);
    SIMPLE_ARRAY_TEST(create_bit,
                      verify_array_with_bit);
}

static void verify_array_with_array_with_bit(verify_data_t *r,
                                             uint32_t i,
                                             void *arg) {
    tsf_bool_t (*values)[3]=arg;
    tsf_type_t *t;
    TAT(t=tsf_type_create(TSF_TK_BIT));
    TAT(t=tsf_type_create_array(t));
    verify_array(t,
                 r,
                 cur_inner_elements,
                 3,
                 verify_array_with_bit,
                 values[i]);
}

static void test_array_with_array_with_bit(uint32_t outer_elements,
                                           uint32_t inner_elements,
                                           ...) {
    tsf_type_t *t=tsf_type_create(TSF_TK_BIT);
    tsf_bool_t values[3][3];
    TAT(t);
    SIMPLE_ARRAY_WITH_ARRAY_TEST(create_bit,
                                 verify_array_with_array_with_bit,
                                 tsf_arg_bool_t);
}

static void verify_array_with_string(verify_data_t *r,
                                     uint32_t i,
                                     void *arg) {
    uint32_t *sizes=arg;
    verify_string(r,sizes[i]);
}

static void test_array_with_string(uint32_t total_elements,
                                   uint32_t value1,
                                   uint32_t value2,
                                   uint32_t value3) {
    tsf_type_t *t=tsf_type_create(TSF_TK_STRING);
    uint32_t values[3];
    TAT(t);
    SIMPLE_ARRAY_TEST(create_string,
                      verify_array_with_string);
}

static void verify_array_with_array_with_string(verify_data_t *r,
                                                uint32_t i,
                                                void *arg) {
    uint32_t (*sizes)[3]=arg;
    tsf_type_t *t;
    TAT(t=tsf_type_create(TSF_TK_STRING));
    TAT(t=tsf_type_create_array(t));
    verify_array(t,
                 r,
                 cur_inner_elements,
                 3,
                 verify_array_with_string,
                 sizes[i]);
}

static void test_array_with_array_with_string(uint32_t outer_elements,
                                              uint32_t inner_elements,
                                              ...) {
    tsf_type_t *t=tsf_type_create(TSF_TK_STRING);
    uint32_t values[3][3];
    TAT(t);
    SIMPLE_ARRAY_WITH_ARRAY_TEST(create_string,
                                 verify_array_with_array_with_string,
                                 uint32_t);
}

static tsf_bool_t current_bit_value;
static uint32_t current_str_size;

static void
verify_array_with_any_with_void_bit_string(verify_data_t *r,
                                           uint32_t i,
                                           void *arg) {
    tsf_buffer_t *buf=verify_any(r);
    switch (i) {
    case 0:
        verify_void(verify_from_buffer(tsf_type_create(TSF_TK_VOID),
                                       buf));
        break;
    case 1:
        verify_bit(verify_from_buffer(tsf_type_create(TSF_TK_BIT),
                                      buf),
                   current_bit_value);
        break;
    case 2:
        verify_string(verify_from_buffer(tsf_type_create(TSF_TK_STRING),
                                         buf),
                      current_str_size);
        break;
    default:
        TF("bad value for i");
    }
}

static void
test_array_with_any_with_void_bit_string(uint32_t total_elements,
                                         tsf_bool_t bit_value,
                                         uint32_t str_size) {
    tsf_type_t *void_type=tsf_type_create(TSF_TK_VOID);
    tsf_type_t *bit_type=tsf_type_create(TSF_TK_BIT);
    tsf_type_t *str_type=tsf_type_create(TSF_TK_STRING);
    tsf_type_t *any_type=tsf_type_create(TSF_TK_ANY);
    tsf_type_t *arr_type;
    create_data_t *ele[3];
    init_pipe();
    TAT(void_type);
    TAT(bit_type);
    TAT(str_type);
    TAT(any_type);
    TAT(arr_type=tsf_type_create_array(any_type));
    current_bit_value=bit_value;
    current_str_size=str_size;
    ele[0]=create_any(create_make_buffer(void_type,
                                         create_void()));
    ele[1]=create_any(create_make_buffer(bit_type,
                                         create_bit(bit_value)));
    ele[2]=create_any(create_make_buffer(str_type,
                                         create_string(str_size)));
    verify_array(arr_type,
                 write_and_read(arr_type,
                                create_array(any_type,
                                             total_elements,
                                             3,
                                             ele)),
                 total_elements,
                 3,
                 verify_array_with_any_with_void_bit_string,
                 NULL);
}

static void verify_struct_with_void(verify_data_t *r,
                                    uint32_t i,
                                    void *arg) {
    verify_void(r);
}

static void test_struct_with_void(const char *name1,
                                  const char *name2,
                                  const char *name3) {
    tsf_type_t *t=tsf_type_create(TSF_TK_VOID);
    tsf_type_t *struct_type;
    create_data_t *ele[3];
    init_pipe();
    TAT(t);
    TAT(struct_type=tsf_type_create_struct(name1,
                                           t,
                                           name2,
                                           t,
                                           name3,
                                           t,
                                           NULL));
    ele[0]=create_void();
    ele[1]=create_void();
    ele[2]=create_void();
    verify_struct(struct_type,
                  write_and_read(struct_type,
                                 create_struct(struct_type,
                                               ele)),
                  verify_struct_with_void,
                  NULL);
}

static tsf_bool_t current_void_enabled;
static tsf_bool_t current_int_enabled;
static tsf_bool_t current_str_enabled;

static int32_t current_int_value;
static uint32_t current_str_size;

static void verify_struct_with_void_int_str_bits(verify_data_t *r,
                                                 uint32_t i,
                                                 void *arg) {
    if (current_void_enabled) {
        if (i==0) {
            verify_void(r);
            return;
        }
        --i;
    }
    if (current_int_enabled) {
        if (i==0) {
            verify_int32_t(r,current_int_value);
            return;
        }
        --i;
    }
    if (current_str_enabled) {
        if (i==0) {
            verify_string(r,current_str_size);
            return;
        }
        --i;
    }
    verify_bit(r,(i%3)==0);
}

static void test_struct_with_void_int_str_bits(tsf_bool_t void_enabled,
                                               tsf_bool_t int_enabled,
                                               tsf_bool_t str_enabled,
                                               int32_t int_value,
                                               uint32_t str_size,
                                               uint32_t num_bits) {
    tsf_type_t *t=tsf_type_create(TSF_TK_BIT);
    tsf_type_t *struct_type;
    create_data_t **ele;
    uint32_t i,pre_size=0;
    init_pipe();
    current_void_enabled=void_enabled;
    current_int_enabled=int_enabled;
    current_str_enabled=str_enabled;
    current_int_value=int_value;
    current_str_size=str_size;
    TAT(t);
    TAT(struct_type=tsf_type_create(TSF_TK_STRUCT));
    TAE(ele=malloc(sizeof(create_data_t*)*
                   (num_bits+
                    (void_enabled?1:0)+
                    (int_enabled?1:0)+
                    (str_enabled?1:0))));
    if (void_enabled) {
        TAT(tsf_struct_type_append(&struct_type,
                                   "void",
                                   tsf_type_create(TSF_TK_VOID)));
        ele[pre_size]=create_void();
        pre_size++;
    }
    if (int_enabled) {
        TAT(tsf_struct_type_append(&struct_type,
                                   "int",
                                   tsf_type_create(TSF_TK_INT32)));
        ele[pre_size]=create_int32_t(int_value);
        pre_size++;
    }
    if (str_enabled) {
        TAT(tsf_struct_type_append(&struct_type,
                                   "str",
                                   tsf_type_create(TSF_TK_STRING)));
        ele[pre_size]=create_string(str_size);
        pre_size++;
    }
    for (i=0;i<num_bits;++i) {
        char buf[64];
        sprintf(buf,"bit" fui32,i);
        TAT(tsf_struct_type_append(&struct_type,
                                   buf,
                                   t));
        ele[i+pre_size]=create_bit((i%3)==0);
    }
    verify_struct(struct_type,
                  write_and_read(struct_type,
                                 create_struct(struct_type,
                                               ele)),
                  verify_struct_with_void_int_str_bits,
                  NULL);
}

static void verify_struct_with_string(verify_data_t *r,
                                      uint32_t i,
                                      void *arg) {
    verify_string(r,((uint32_t*)arg)[i]);
}

static void test_struct_with_string(const char *name1,
                                    uint32_t value1,
                                    const char *name2,
                                    uint32_t value2,
                                    const char *name3,
                                    uint32_t value3) {
    tsf_type_t *t=tsf_type_create(TSF_TK_STRING);
    uint32_t values[3];
    TAT(t);
    SIMPLE_STRUCT_TEST(create_string,
                       verify_struct_with_string);
}

static void verify_struct_with_any_with_void(verify_data_t *r,
                                             uint32_t i,
                                             void *arg) {
    verify_void(verify_from_buffer(tsf_type_create(TSF_TK_VOID),
                                   verify_any(r)));
}

static void test_struct_with_any_with_void(const char *name1,
                                           const char *name2,
                                           const char *name3) {
    tsf_type_t *void_type=tsf_type_create(TSF_TK_VOID);
    tsf_type_t *struct_type=
        tsf_type_create_struct(name1,
                               tsf_type_create(TSF_TK_ANY),
                               name2,
                               tsf_type_create(TSF_TK_ANY),
                               name3,
                               tsf_type_create(TSF_TK_ANY),
                               NULL);
    create_data_t *ele[3];
    init_pipe();
    TAT(void_type);
    TAT(struct_type);
    ele[0]=create_any(create_make_buffer(void_type,
                                         create_void()));
    ele[1]=create_any(create_make_buffer(void_type,
                                         create_void()));
    ele[2]=create_any(create_make_buffer(void_type,
                                         create_void()));
    verify_struct(struct_type,
                  write_and_read(struct_type,
                                 create_struct(struct_type,
                                               ele)),
                  verify_struct_with_any_with_void,
                  NULL);
}

static void test_any_with_struct_with_uint8_t(const char *name1,
                                              uint8_t value1,
                                              const char *name2,
                                              uint8_t value2,
                                              const char *name3,
                                              uint8_t value3) {
    tsf_type_t *t=tsf_type_create(TSF_TK_UINT8);
    tsf_type_t *any_t=tsf_type_create(TSF_TK_ANY);
    tsf_type_t *struct_type=tsf_type_create_struct(name1,
                                                   t,
                                                   name2,
                                                   t,
                                                   name3,
                                                   t,
                                                   NULL);
    uint8_t values[3];
    create_data_t *ele[3];
    tsf_buffer_t *buf;
    TAT(t);
    TAT(struct_type);
    TAT(any_t);
    init_pipe();
    ele[0]=create_uint8_t(values[0]=value1);
    ele[1]=create_uint8_t(values[1]=value2);
    ele[2]=create_uint8_t(values[2]=value3);
    buf=create_make_buffer(struct_type,
                           create_struct(struct_type,
                                         ele));
    buf=verify_any(write_and_read(any_t,
                                  create_any(buf)));
    verify_struct(struct_type,
                  verify_from_buffer(struct_type,buf),
                  verify_struct_with_uint8_t,
                  values);
}

static uint32_t array_sizes[3];

#define DEFINE_STRUCT_WITH_ARRAY_TEST(utype,ltype,ctype,arg_t)          \
    static void                                                         \
    verify_struct_with_array_with_##ltype(verify_data_t *r,             \
                                          uint32_t i,                   \
                                          void *arg) {                  \
        ctype (*values)[3]=arg;                                         \
        verify_array(tsf_type_create_array(tsf_type_create(utype)),     \
                     r,                                                 \
                     array_sizes[i],                                    \
                     3,                                                 \
                     verify_array_with_##ltype,                         \
                     values[i]);                                        \
    }                                                                   \
                                                                        \
    static void                                                         \
    test_struct_with_array_with_##ltype(const char *name1,              \
                                        uint32_t num_elements1,         \
                                        const char *name2,              \
                                        uint32_t num_elements2,         \
                                        const char *name3,              \
                                        uint32_t num_elements3,         \
                                        ...) {                          \
        tsf_type_t *t,*arr_t,*struct_t;                                 \
        va_list lst;                                                    \
        create_data_t *ele[3];                                          \
        ctype values[3][3];                                             \
        uint32_t i,j;                                                   \
        TAT(t=tsf_type_create(utype));                                  \
        TAT(arr_t=tsf_type_create_array(t));                            \
        TAT(struct_t=tsf_type_create_struct(name1,arr_t,                \
                                            name2,arr_t,                \
                                            name3,arr_t,                \
                                            NULL));                     \
        array_sizes[0]=num_elements1;                                   \
        array_sizes[1]=num_elements2;                                   \
        array_sizes[2]=num_elements3;                                   \
        init_pipe();                                                    \
        va_start(lst,num_elements3);                                    \
        for (i=0;i<3;++i) {                                             \
            create_data_t *sub_ele[3];                                  \
            for (j=0;j<3;++j) {                                         \
                sub_ele[j]=create_##ltype                               \
                    (values[i][j]=va_arg(lst,arg_t));                   \
            }                                                           \
            ele[i]=create_array(t,                                      \
                                array_sizes[i],                         \
                                3,                                      \
                                sub_ele);                               \
        }                                                               \
        va_end(lst);                                                    \
        verify_struct(struct_t,                                         \
                      write_and_read(struct_t,                          \
                                     create_struct(struct_t,            \
                                                   ele)),               \
                      verify_struct_with_array_with_##ltype,            \
                      values);                                          \
    }                                                                   \
    typedef void define_struct_with_array_test_##utype

DEFINE_STRUCT_WITH_ARRAY_TEST(TSF_TK_UINT8,uint8_t,uint8_t,unsigned);
DEFINE_STRUCT_WITH_ARRAY_TEST(TSF_TK_BIT,bit,tsf_bool_t,tsf_arg_bool_t);

static tsf_type_t *current_struct_type;

static void verify_array_with_struct_with_float(verify_data_t *r,
                                                uint32_t i,
                                                void *arg) {
    float (*values)[3]=arg;
    verify_struct(current_struct_type,
                  r,
                  verify_struct_with_float,
                  values[i]);
}

static void test_array_with_struct_with_float(uint32_t num_elements,
                                              const char *name1,
                                              const char *name2,
                                              const char *name3,
                                              ...) {
    tsf_type_t *t;
    tsf_type_t *arr_t;
    float values[3][3];
    va_list lst;
    uint32_t i,j;
    create_data_t *ele[3];
    TAT(t=tsf_type_create(TSF_TK_FLOAT));
    TAT(current_struct_type=tsf_type_create_struct(name1,t,
                                                   name2,t,
                                                   name3,t,
                                                   NULL));
    TAT(arr_t=tsf_type_create_array(current_struct_type));
    init_pipe();
    va_start(lst,name3);
    for (i=0;i<3;++i) {
        create_data_t *sub_ele[3];
        for (j=0;j<3;++j) {
            sub_ele[j]=create_float(values[i][j]=va_arg(lst,double));
        }
        ele[i]=create_struct(current_struct_type,
                             sub_ele);
    }
    va_end(lst);
    verify_array(arr_t,
                 write_and_read(arr_t,
                                create_array(current_struct_type,
                                             num_elements,
                                             3,
                                             ele)),
                 num_elements,
                 3,
                 verify_array_with_struct_with_float,
                 values);
}

tsf_type_t *inner_struct_types[3];

static void verify_struct_with_struct_with_string(verify_data_t *r,
                                                  uint32_t i,
                                                  void *arg) {
    uint32_t (*sizes)[3]=arg;
    verify_struct(inner_struct_types[i],
                  r,
                  verify_struct_with_string,
                  sizes[i]);
}

static void test_struct_with_struct_with_string(const char *top_name1,
                                                const char *top_name2,
                                                const char *top_name3,
                                                ...) {
    tsf_type_t *t,*outer_struct_type;
    create_data_t *ele[3];
    uint32_t sizes[3][3];
    uint32_t i,j;
    va_list lst;
    const char *top_names[3];
    top_names[0]=top_name1;
    top_names[1]=top_name2;
    top_names[2]=top_name3;
    TAT(t=tsf_type_create(TSF_TK_STRING));
    TAT(outer_struct_type=tsf_type_create(TSF_TK_STRUCT));
    va_start(lst,top_name3);
    for (i=0;i<3;++i) {
        create_data_t *sub_ele[3];
        TAT(inner_struct_types[i]=tsf_type_create(TSF_TK_STRUCT));
        for (j=0;j<3;++j) {
            const char *name=va_arg(lst,const char*);
            sizes[i][j]=va_arg(lst,uint32_t);
            TAT(tsf_struct_type_append(inner_struct_types+i,
                                       name,
                                       t));
            sub_ele[j]=create_string(sizes[i][j]);
        }
        TAT(tsf_struct_type_append(&outer_struct_type,
                                   top_names[i],
                                   inner_struct_types[i]));
        ele[i]=create_struct(inner_struct_types[i],
                             sub_ele);
    }
    va_end(lst);
    init_pipe();
    verify_struct(outer_struct_type,
                  write_and_read(outer_struct_type,
                                 create_struct(outer_struct_type,
                                               ele)),
                  verify_struct_with_struct_with_string,
                  sizes);
}

static void test_choice_with_void(uint32_t value,
                                  ...) {
    va_list lst;
    tsf_type_t *t,*choice_t;
    TAT(t=tsf_type_create(TSF_TK_VOID));
    TAT(choice_t=tsf_type_create(TSF_TK_CHOICE));
    va_start(lst,value);
    for (;;) {
        const char *name=va_arg(lst,const char*);
        if (name==NULL) {
            break;
        }
        TAT(tsf_choice_type_append(&choice_t,
                                   name,
                                   t));
    }
    init_pipe();
    verify_choice(choice_t,
                  write_and_read(choice_t,
                                 create_choice(choice_t,
                                               value,
                                               create_void())),
                  value);
}

static void test_choice_with_bit(const char *name1,
                                 const char *name2,
                                 const char *name3,
                                 uint32_t value,
                                 tsf_bool_t data) {
    tsf_type_t *t;
    TAT(t=tsf_type_create(TSF_TK_BIT));
    SIMPLE_CHOICE_TEST(create_bit,
                       verify_bit);
}

static void test_choice_with_string(const char *name1,
                                    const char *name2,
                                    const char *name3,
                                    uint32_t value,
                                    uint32_t data) {
    tsf_type_t *t;
    TAT(t=tsf_type_create(TSF_TK_STRING));
    SIMPLE_CHOICE_TEST(create_string,
                       verify_string);
}

static tsf_type_t *current_choice_t;
static uint32_t current_num_void_elems;

static void verify_struct_with_choice(verify_data_t *r,
                                      uint32_t i,
                                      void *arg) {
    uint32_t value=*((uint32_t*)arg);
    switch (i) {
    case 0: /* the choice */
        r=verify_choice(current_choice_t,
                        r,
                        value);
        if (value >= current_num_void_elems &&
            value != UINT32_MAX) {
            verify_uint32_t(r,value*7);
        }
        break;
    case 1: /* the suffix */
        verify_uint32_t(r,value*19);
        break;
    default:
        TF("bad struct element index");
    }
}

static void test_struct_with_choice(uint32_t num_choice_elems,
                                    uint32_t num_void_elems,
                                    uint32_t value) {
    tsf_type_t *int_t,*void_t;
    tsf_type_t *choice_t,*struct_t;
    uint32_t i;
    create_data_t *cr=NULL;
    create_data_t *ele[2];
    current_num_void_elems=num_void_elems;
    TAT(int_t=tsf_type_create(TSF_TK_UINT32));
    TAT(void_t=tsf_type_create(TSF_TK_VOID));
    TAT(choice_t=tsf_type_create(TSF_TK_CHOICE));
    for (i=0;i<num_choice_elems;++i) {
        char buf[64];
        sprintf(buf,"ele" fui32,i);
        TAT(tsf_choice_type_append(&choice_t,
                                   buf,
                                   (i<num_void_elems?
                                    void_t:int_t)));
    }
    current_choice_t=choice_t;
    TAT(struct_t=tsf_type_create_struct("choice",choice_t,
                                        "suffix",int_t,
                                        NULL));
    init_pipe();
    if (value!=UINT32_MAX) {
        if (value<num_void_elems) {
            cr=create_void();
        } else {
            cr=create_uint32_t(value*7);
        }
    }
    ele[0]=create_choice(choice_t,
                         value,
                         cr);
    ele[1]=create_uint32_t(value*19);
    verify_struct(struct_t,
                  write_and_read(struct_t,
                                 create_struct(struct_t,
                                               ele)),
                  verify_struct_with_choice,
                  &value);
}

static void do_all_tests() {
    DO_TEST(test_int8_t(INT8_MIN));
    DO_TEST(test_int8_t(0));
    DO_TEST(test_int8_t(INT8_MAX));
    DO_TEST(test_uint8_t(0));
    DO_TEST(test_uint8_t(UINT8_MAX));
    DO_TEST(test_int16_t(INT16_MIN));
    DO_TEST(test_int16_t(0));
    DO_TEST(test_int16_t(INT16_MAX));
    DO_TEST(test_uint16_t(0));
    DO_TEST(test_uint16_t(UINT16_MAX));
    DO_TEST(test_int32_t(INT32_MIN));
    DO_TEST(test_int32_t(0));
    DO_TEST(test_int32_t(INT32_MAX));
    DO_TEST(test_uint32_t(0));
    DO_TEST(test_uint32_t(UINT32_MAX));
    DO_TEST(test_int64_t(INT64_MIN));
    DO_TEST(test_int64_t(0));
    DO_TEST(test_int64_t(INT64_MAX));
    DO_TEST(test_uint64_t(0));
    DO_TEST(test_uint64_t(UINT64_MAX));
    DO_TEST(test_float(0.0));
    DO_TEST(test_float(-5.0));
    DO_TEST(test_float(5.0));
    DO_TEST(test_double(0.0));
    DO_TEST(test_double(-10000.0));
    DO_TEST(test_double(10000.0));

    DO_TEST(test_any_with_int8_t(INT8_MIN));
    DO_TEST(test_any_with_int8_t(0));
    DO_TEST(test_any_with_int8_t(INT8_MAX));
    DO_TEST(test_any_with_uint8_t(0));
    DO_TEST(test_any_with_uint8_t(UINT8_MAX));
    DO_TEST(test_any_with_int16_t(INT16_MIN));
    DO_TEST(test_any_with_int16_t(0));
    DO_TEST(test_any_with_int16_t(INT16_MAX));
    DO_TEST(test_any_with_uint16_t(0));
    DO_TEST(test_any_with_uint16_t(UINT16_MAX));
    DO_TEST(test_any_with_int32_t(INT32_MIN));
    DO_TEST(test_any_with_int32_t(0));
    DO_TEST(test_any_with_int32_t(INT32_MAX));
    DO_TEST(test_any_with_uint32_t(0));
    DO_TEST(test_any_with_uint32_t(UINT32_MAX));
    DO_TEST(test_any_with_int64_t(INT64_MIN));
    DO_TEST(test_any_with_int64_t(0));
    DO_TEST(test_any_with_int64_t(INT64_MAX));
    DO_TEST(test_any_with_uint64_t(0));
    DO_TEST(test_any_with_uint64_t(UINT64_MAX));
    DO_TEST(test_any_with_float(0.0));
    DO_TEST(test_any_with_float(-5.0));
    DO_TEST(test_any_with_float(5.0));
    DO_TEST(test_any_with_double(0.0));
    DO_TEST(test_any_with_double(-10000.0));
    DO_TEST(test_any_with_double(10000.0));

    DO_TEST(test_array_with_int8_t(0,-5,0,5));
    DO_TEST(test_array_with_int8_t(1,-5,0,5));
    DO_TEST(test_array_with_int8_t(5,-5,0,5));

    DO_TEST(test_array_with_uint8_t(0,0,5,10));
    DO_TEST(test_array_with_uint8_t(1,0,5,10));
    DO_TEST(test_array_with_uint8_t(5,0,5,10));

    DO_TEST(test_array_with_int16_t(0,-500,0,500));
    DO_TEST(test_array_with_int16_t(1,-500,0,500));
    DO_TEST(test_array_with_int16_t(5,-500,0,500));

    DO_TEST(test_array_with_uint16_t(0,0,500,1000));
    DO_TEST(test_array_with_uint16_t(1,0,500,1000));
    DO_TEST(test_array_with_uint16_t(5,0,500,1000));

    DO_TEST(test_array_with_int32_t(0,-5000000,0,5000000));
    DO_TEST(test_array_with_int32_t(1,-5000000,0,5000000));
    DO_TEST(test_array_with_int32_t(5,-5000000,0,5000000));

    DO_TEST(test_array_with_uint32_t(0,0,5000000,10000000));
    DO_TEST(test_array_with_uint32_t(1,0,5000000,10000000));
    DO_TEST(test_array_with_uint32_t(5,0,5000000,10000000));

    DO_TEST(test_array_with_int64_t(0,
                                    INT64_C(-50000000000),
                                    INT64_C(0),
                                    INT64_C(50000000000)));
    DO_TEST(test_array_with_int64_t(1,
                                    INT64_C(-50000000000),
                                    INT64_C(0),
                                    INT64_C(50000000000)));
    DO_TEST(test_array_with_int64_t(5,
                                    INT64_C(-50000000000),
                                    INT64_C(0),
                                    INT64_C(50000000000)));

    DO_TEST(test_array_with_uint64_t(0,
                                     INT64_C(0),
                                     INT64_C(50000000000),
                                     INT64_C(100000000000)));
    DO_TEST(test_array_with_uint64_t(1,
                                     INT64_C(0),
                                     INT64_C(50000000000),
                                     INT64_C(100000000000)));
    DO_TEST(test_array_with_uint64_t(5,
                                     INT64_C(0),
                                     INT64_C(50000000000),
                                     INT64_C(100000000000)));

    DO_TEST(test_array_with_float(0,-500.0,0.0,500.0));
    DO_TEST(test_array_with_float(1,-500.0,0.0,500.0));
    DO_TEST(test_array_with_float(5,-500.0,0.0,500.0));

    DO_TEST(test_array_with_double(0,-500.0,0.0,500.0));
    DO_TEST(test_array_with_double(1,-500.0,0.0,500.0));
    DO_TEST(test_array_with_double(5,-500.0,0.0,500.0));

    DO_TEST(test_array_with_array_with_int8_t(1,0,-4,-3,-2,-1,0,1,2,3,4));
    DO_TEST(test_array_with_array_with_int8_t(1,1,-4,-3,-2,-1,0,1,2,3,4));
    DO_TEST(test_array_with_array_with_int8_t(5,1,-4,-3,-2,-1,0,1,2,3,4));
    DO_TEST(test_array_with_array_with_int8_t(1,5,-4,-3,-2,-1,0,1,2,3,4));
    DO_TEST(test_array_with_array_with_int8_t(5,5,-4,-3,-2,-1,0,1,2,3,4));

    DO_TEST(test_array_with_array_with_uint8_t(1,0,0,1,2,3,4,5,6,7,8));
    DO_TEST(test_array_with_array_with_uint8_t(1,1,0,1,2,3,4,5,6,7,8));
    DO_TEST(test_array_with_array_with_uint8_t(1,5,0,1,2,3,4,5,6,7,8));
    DO_TEST(test_array_with_array_with_uint8_t(5,1,0,1,2,3,4,5,6,7,8));
    DO_TEST(test_array_with_array_with_uint8_t(5,5,0,1,2,3,4,5,6,7,8));

    DO_TEST(test_array_with_array_with_int16_t(1,0,-400,-300,-200,-100,0,100,200,300,400));
    DO_TEST(test_array_with_array_with_int16_t(1,1,-400,-300,-200,-100,0,100,200,300,400));
    DO_TEST(test_array_with_array_with_int16_t(1,5,-400,-300,-200,-100,0,100,200,300,400));
    DO_TEST(test_array_with_array_with_int16_t(5,1,-400,-300,-200,-100,0,100,200,300,400));
    DO_TEST(test_array_with_array_with_int16_t(5,5,-400,-300,-200,-100,0,100,200,300,400));

    DO_TEST(test_array_with_array_with_uint16_t(1,0,0,100,200,300,400,500,600,700,800));
    DO_TEST(test_array_with_array_with_uint16_t(1,1,0,100,200,300,400,500,600,700,800));
    DO_TEST(test_array_with_array_with_uint16_t(1,5,0,100,200,300,400,500,600,700,800));
    DO_TEST(test_array_with_array_with_uint16_t(5,1,0,100,200,300,400,500,600,700,800));
    DO_TEST(test_array_with_array_with_uint16_t(5,5,0,100,200,300,400,500,600,700,800));

    DO_TEST(test_array_with_array_with_int32_t(1,0,-4000000,-3000000,-2000000,-1000000,0,1000000,2000000,3000000,4000000));
    DO_TEST(test_array_with_array_with_int32_t(1,1,-4000000,-3000000,-2000000,-1000000,0,1000000,2000000,3000000,4000000));
    DO_TEST(test_array_with_array_with_int32_t(1,5,-4000000,-3000000,-2000000,-1000000,0,1000000,2000000,3000000,4000000));
    DO_TEST(test_array_with_array_with_int32_t(5,1,-4000000,-3000000,-2000000,-1000000,0,1000000,2000000,3000000,4000000));
    DO_TEST(test_array_with_array_with_int32_t(5,5,-4000000,-3000000,-2000000,-1000000,0,1000000,2000000,3000000,4000000));

    DO_TEST(test_array_with_array_with_uint32_t(1,0,0,1000000,2000000,3000000,4000000,5000000,6000000,7000000,8000000));
    DO_TEST(test_array_with_array_with_uint32_t(1,1,0,1000000,2000000,3000000,4000000,5000000,6000000,7000000,8000000));
    DO_TEST(test_array_with_array_with_uint32_t(1,5,0,1000000,2000000,3000000,4000000,5000000,6000000,7000000,8000000));
    DO_TEST(test_array_with_array_with_uint32_t(5,1,0,1000000,2000000,3000000,4000000,5000000,6000000,7000000,8000000));
    DO_TEST(test_array_with_array_with_uint32_t(5,5,0,1000000,2000000,3000000,4000000,5000000,6000000,7000000,8000000));

    DO_TEST(test_array_with_array_with_int64_t(1,0,INT64_C(-40000000000),INT64_C(-30000000000),INT64_C(-20000000000),INT64_C(-10000000000),INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000)));
    DO_TEST(test_array_with_array_with_int64_t(1,1,INT64_C(-40000000000),INT64_C(-30000000000),INT64_C(-20000000000),INT64_C(-10000000000),INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000)));
    DO_TEST(test_array_with_array_with_int64_t(5,1,INT64_C(-40000000000),INT64_C(-30000000000),INT64_C(-20000000000),INT64_C(-10000000000),INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000)));
    DO_TEST(test_array_with_array_with_int64_t(1,5,INT64_C(-40000000000),INT64_C(-30000000000),INT64_C(-20000000000),INT64_C(-10000000000),INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000)));
    DO_TEST(test_array_with_array_with_int64_t(5,5,INT64_C(-40000000000),INT64_C(-30000000000),INT64_C(-20000000000),INT64_C(-10000000000),INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000)));

    DO_TEST(test_array_with_array_with_uint64_t(1,0,INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000),INT64_C(50000000000),INT64_C(60000000000),INT64_C(70000000000),INT64_C(80000000000)));
    DO_TEST(test_array_with_array_with_uint64_t(1,1,INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000),INT64_C(50000000000),INT64_C(60000000000),INT64_C(70000000000),INT64_C(80000000000)));
    DO_TEST(test_array_with_array_with_uint64_t(1,5,INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000),INT64_C(50000000000),INT64_C(60000000000),INT64_C(70000000000),INT64_C(80000000000)));
    DO_TEST(test_array_with_array_with_uint64_t(5,1,INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000),INT64_C(50000000000),INT64_C(60000000000),INT64_C(70000000000),INT64_C(80000000000)));
    DO_TEST(test_array_with_array_with_uint64_t(5,5,INT64_C(0),INT64_C(10000000000),INT64_C(20000000000),INT64_C(30000000000),INT64_C(40000000000),INT64_C(50000000000),INT64_C(60000000000),INT64_C(70000000000),INT64_C(80000000000)));

    DO_TEST(test_array_with_array_with_float(1,0,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_float(1,1,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_float(5,1,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_float(1,5,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_float(5,5,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));

    DO_TEST(test_array_with_array_with_double(1,0,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_double(1,1,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_double(5,1,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_double(1,5,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));
    DO_TEST(test_array_with_array_with_double(5,5,-4.0,-3.0,-2.0,-1.0,0.0,1.0,2.0,3.0,4.0));

    DO_TEST(test_struct_with_int8_t("the first one",-5,"the second one",0,"the third one",5));
    DO_TEST(test_struct_with_uint8_t("the first one",0,"the second one",5,"the third one",10));
    DO_TEST(test_struct_with_int16_t("the first one",-500,"the second one",0,"the third one",500));
    DO_TEST(test_struct_with_uint16_t("the first one",0,"the second one",500,"the third one",1000));
    DO_TEST(test_struct_with_int32_t("the first one",-5000000,"the second one",0,"the third one",5000000));
    DO_TEST(test_struct_with_uint32_t("the first one",0,"the second one",5000000,"the third one",10000000));
    DO_TEST(test_struct_with_int64_t("the first one",INT64_C(-50000000000),
                                     "the second one",INT64_C(0),
                                     "the third one",INT64_C(50000000000)));
    DO_TEST(test_struct_with_uint64_t("the first one",INT64_C(0),
                                      "the second one",INT64_C(50000000000),
                                      "the third one",INT64_C(100000000000)));
    DO_TEST(test_struct_with_float("the first one",-500.0,"the second one",0.0,"the third one",500.0));
    DO_TEST(test_struct_with_double("the first one",-500.0,"the second one",0.0,"the third one",500.0));
    
    DO_TEST(test_choice_with_int8_t("the first one","the second one","the third one",0,-5));
    DO_TEST(test_choice_with_int8_t("the first one","the second one","the third one",1,-5));
    DO_TEST(test_choice_with_int8_t("the first one","the second one","the third one",2,-5));
    DO_TEST(test_choice_with_int8_t("the first one","the second one","the third one",UINT32_MAX,-5));
    
    DO_TEST(test_choice_with_uint8_t("the first one","the second one","the third one",0,5));
    DO_TEST(test_choice_with_uint8_t("the first one","the second one","the third one",1,5));
    DO_TEST(test_choice_with_uint8_t("the first one","the second one","the third one",2,5));
    DO_TEST(test_choice_with_uint8_t("the first one","the second one","the third one",UINT32_MAX,5));
    
    DO_TEST(test_choice_with_int16_t("the first one","the second one","the third one",0,-500));
    DO_TEST(test_choice_with_int16_t("the first one","the second one","the third one",1,-500));
    DO_TEST(test_choice_with_int16_t("the first one","the second one","the third one",2,-500));
    DO_TEST(test_choice_with_int16_t("the first one","the second one","the third one",UINT32_MAX,-500));
    
    DO_TEST(test_choice_with_uint16_t("the first one","the second one","the third one",0,500));
    DO_TEST(test_choice_with_uint16_t("the first one","the second one","the third one",1,500));
    DO_TEST(test_choice_with_uint16_t("the first one","the second one","the third one",2,500));
    DO_TEST(test_choice_with_uint16_t("the first one","the second one","the third one",UINT32_MAX,500));
    
    DO_TEST(test_choice_with_int32_t("the first one","the second one","the third one",0,-5000000));
    DO_TEST(test_choice_with_int32_t("the first one","the second one","the third one",1,-5000000));
    DO_TEST(test_choice_with_int32_t("the first one","the second one","the third one",2,-5000000));
    DO_TEST(test_choice_with_int32_t("the first one","the second one","the third one",UINT32_MAX,-5000000));
    
    DO_TEST(test_choice_with_uint32_t("the first one","the second one","the third one",0,5000000));
    DO_TEST(test_choice_with_uint32_t("the first one","the second one","the third one",1,5000000));
    DO_TEST(test_choice_with_uint32_t("the first one","the second one","the third one",2,5000000));
    DO_TEST(test_choice_with_uint32_t("the first one","the second one","the third one",UINT32_MAX,5000000));
    
    DO_TEST(test_choice_with_int64_t("the first one","the second one","the third one",0,INT64_C(-50000000000)));
    DO_TEST(test_choice_with_int64_t("the first one","the second one","the third one",1,INT64_C(-50000000000)));
    DO_TEST(test_choice_with_int64_t("the first one","the second one","the third one",2,INT64_C(-50000000000)));
    DO_TEST(test_choice_with_int64_t("the first one","the second one","the third one",UINT32_MAX,INT64_C(-50000000000)));
    
    DO_TEST(test_choice_with_uint64_t("the first one","the second one","the third one",0,UINT64_C(50000000000)));
    DO_TEST(test_choice_with_uint64_t("the first one","the second one","the third one",1,UINT64_C(50000000000)));
    DO_TEST(test_choice_with_uint64_t("the first one","the second one","the third one",2,UINT64_C(50000000000)));
    DO_TEST(test_choice_with_uint64_t("the first one","the second one","the third one",UINT32_MAX,UINT64_C(50000000000)));

    DO_TEST(test_choice_with_float("a","b","c",0,-5.0));
    DO_TEST(test_choice_with_float("a","b","c",1,-5.0));
    DO_TEST(test_choice_with_float("a","b","c",2,-5.0));
    DO_TEST(test_choice_with_float("a","b","c",UINT32_MAX,-5.0));
    
    DO_TEST(test_choice_with_double("a","b","c",0,-5.0));
    DO_TEST(test_choice_with_double("a","b","c",1,-5.0));
    DO_TEST(test_choice_with_double("a","b","c",2,-5.0));
    DO_TEST(test_choice_with_double("a","b","c",UINT32_MAX,-5.0));
    
    DO_TEST(test_bit(tsf_false));
    DO_TEST(test_bit(tsf_true));
    
    DO_TEST(test_void());
    
    DO_TEST(test_string(0));
    DO_TEST(test_string(1));
    DO_TEST(test_string(5));
    DO_TEST(test_string(126));
    DO_TEST(test_string(127));
    DO_TEST(test_string(128));
    DO_TEST(test_string(129));
    DO_TEST(test_string(254));
    DO_TEST(test_string(255));
    DO_TEST(test_string(256));
    DO_TEST(test_string(257));
    DO_TEST(test_string(TSF_REGION_SIZE_UNIT-2));
    DO_TEST(test_string(TSF_REGION_SIZE_UNIT-1));
    DO_TEST(test_string(TSF_REGION_SIZE_UNIT));
    DO_TEST(test_string(TSF_REGION_SIZE_UNIT+1));
    DO_TEST(test_string(2*TSF_REGION_SIZE_UNIT-2));
    DO_TEST(test_string(2*TSF_REGION_SIZE_UNIT-1));
    DO_TEST(test_string(2*TSF_REGION_SIZE_UNIT));
    DO_TEST(test_string(2*TSF_REGION_SIZE_UNIT+1));
    
    DO_TEST(test_any_with_void());
    
    DO_TEST(test_any_with_bit(tsf_false));
    DO_TEST(test_any_with_bit(tsf_true));
    
    DO_TEST(test_any_with_string(0));
    DO_TEST(test_any_with_string(1));
    DO_TEST(test_any_with_string(5));
    DO_TEST(test_any_with_string(254));
    DO_TEST(test_any_with_string(255));
    DO_TEST(test_any_with_string(256));
    
    DO_TEST(test_array_with_void(0));
    DO_TEST(test_array_with_void(1));
    DO_TEST(test_array_with_void(254));
    DO_TEST(test_array_with_void(255));
    DO_TEST(test_array_with_void(256));
    
    DO_TEST(test_array_with_array_with_void(0,0));
    DO_TEST(test_array_with_array_with_void(1,0));
    DO_TEST(test_array_with_array_with_void(254,0));
    DO_TEST(test_array_with_array_with_void(255,0));
    DO_TEST(test_array_with_array_with_void(256,0));
    DO_TEST(test_array_with_array_with_void(1,1));
    DO_TEST(test_array_with_array_with_void(254,1));
    DO_TEST(test_array_with_array_with_void(255,1));
    DO_TEST(test_array_with_array_with_void(256,1));
    DO_TEST(test_array_with_array_with_void(1,254));
    DO_TEST(test_array_with_array_with_void(254,254));
    DO_TEST(test_array_with_array_with_void(255,254));
    DO_TEST(test_array_with_array_with_void(256,254));
    DO_TEST(test_array_with_array_with_void(1,255));
    DO_TEST(test_array_with_array_with_void(254,255));
    DO_TEST(test_array_with_array_with_void(255,255));
    DO_TEST(test_array_with_array_with_void(256,255));
    DO_TEST(test_array_with_array_with_void(1,256));
    DO_TEST(test_array_with_array_with_void(254,256));
    DO_TEST(test_array_with_array_with_void(255,256));
    DO_TEST(test_array_with_array_with_void(256,256));
    
    DO_TEST(test_array_with_bit(0,tsf_true,tsf_false,tsf_false));
    DO_TEST(test_array_with_bit(1,tsf_true,tsf_false,tsf_false));
    DO_TEST(test_array_with_bit(5,tsf_true,tsf_false,tsf_false));
    
    DO_TEST(test_array_with_array_with_bit(0,0,
                                           tsf_true,tsf_false,tsf_false,
                                           tsf_false,tsf_true,tsf_false,
                                           tsf_false,tsf_false,tsf_true));
    DO_TEST(test_array_with_array_with_bit(1,1,
                                           tsf_true,tsf_false,tsf_false,
                                           tsf_false,tsf_true,tsf_false,
                                           tsf_false,tsf_false,tsf_true));
    DO_TEST(test_array_with_array_with_bit(1,5,
                                           tsf_true,tsf_false,tsf_false,
                                           tsf_false,tsf_true,tsf_false,
                                           tsf_false,tsf_false,tsf_true));
    DO_TEST(test_array_with_array_with_bit(5,1,
                                           tsf_true,tsf_false,tsf_false,
                                           tsf_false,tsf_true,tsf_false,
                                           tsf_false,tsf_false,tsf_true));
    DO_TEST(test_array_with_array_with_bit(5,5,
                                           tsf_true,tsf_false,tsf_false,
                                           tsf_false,tsf_true,tsf_false,
                                           tsf_false,tsf_false,tsf_true));
    
    DO_TEST(test_array_with_string(0,0,10,50));
    DO_TEST(test_array_with_string(1,0,10,50));
    DO_TEST(test_array_with_string(5,0,10,50));
    
    DO_TEST(test_array_with_array_with_string(1,0,0,10,50,60,70,90,5,15,25));
    DO_TEST(test_array_with_array_with_string(1,1,0,10,50,60,70,90,5,15,25));
    DO_TEST(test_array_with_array_with_string(5,1,0,10,50,60,70,90,5,15,25));
    DO_TEST(test_array_with_array_with_string(1,5,0,10,50,60,70,90,5,15,25));
    DO_TEST(test_array_with_array_with_string(5,5,0,10,50,60,70,90,5,15,25));
    
    DO_TEST(test_array_with_any_with_void_bit_string(0,tsf_true,10));
    DO_TEST(test_array_with_any_with_void_bit_string(1,tsf_true,10));
    DO_TEST(test_array_with_any_with_void_bit_string(1,tsf_false,15));
    DO_TEST(test_array_with_any_with_void_bit_string(5,tsf_true,10));
    DO_TEST(test_array_with_any_with_void_bit_string(5,tsf_false,15));
    
    DO_TEST(test_struct_with_void("a","b","c"));
    DO_TEST(test_struct_with_void("","b","c"));
    
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_false,
                                               0,0,0));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_false,
                                               0,0,1));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_false,
                                               0,0,7));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_false,
                                               0,0,8));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_false,
                                               0,0,15));
    
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_false,tsf_false,
                                               0,0,0));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_false,tsf_false,
                                               0,0,1));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_false,tsf_false,
                                               0,0,7));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_false,tsf_false,
                                               0,0,8));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_false,tsf_false,
                                               0,0,15));
    
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_true,tsf_false,
                                               -5,0,0));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_true,tsf_false,
                                               -5,0,1));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_true,tsf_false,
                                               -5,0,7));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_true,tsf_false,
                                               -5,0,8));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_true,tsf_false,
                                               -5,0,15));
    
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_true,
                                               0,25,0));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_true,
                                               0,25,1));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_true,
                                               0,25,7));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_true,
                                               0,25,8));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_false,tsf_false,tsf_true,
                                               0,25,15));
    
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_true,tsf_true,
                                               -7,7,0));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_true,tsf_true,
                                               -7,7,1));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_true,tsf_true,
                                               -7,7,7));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_true,tsf_true,
                                               -7,7,8));
    DO_TEST(test_struct_with_void_int_str_bits(tsf_true,tsf_true,tsf_true,
                                               -7,7,15));
                                               
    DO_TEST(test_struct_with_string("a",0,"b",0,"c",0));
    DO_TEST(test_struct_with_string("a",5,"b",10,"c",15));
    
    DO_TEST(test_struct_with_any_with_void("blah","bleh","blih"));
    
    DO_TEST(test_any_with_struct_with_uint8_t("a",0,"b",1,"c",2));
    
    DO_TEST(test_struct_with_array_with_uint8_t("a",0,"b",0,"c",0,
                                                1,2,3,4,5,6,7,8,9));
    DO_TEST(test_struct_with_array_with_uint8_t("a",1,"b",2,"c",3,
                                                1,2,3,4,5,6,7,8,9));
    DO_TEST(test_struct_with_array_with_uint8_t("a",254,"b",255,"c",256,
                                                1,2,3,4,5,6,7,8,9));
    DO_TEST(test_struct_with_array_with_uint8_t("a",256,"b",255,"c",254,
                                                1,2,3,4,5,6,7,8,9));
    DO_TEST(test_struct_with_array_with_uint8_t("a",255,"b",256,"c",255,
                                                1,2,3,4,5,6,7,8,9));
    DO_TEST(test_struct_with_array_with_uint8_t("a",255,"b",254,"c",255,
                                                1,2,3,4,5,6,7,8,9));
    
    DO_TEST(test_struct_with_array_with_bit("a",0,"b",0,"c",0,
                                            tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false));
    DO_TEST(test_struct_with_array_with_bit("a",1,"b",2,"c",3,
                                            tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false));
    DO_TEST(test_struct_with_array_with_bit("a",254,"b",255,"c",256,
                                            tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false));
    DO_TEST(test_struct_with_array_with_bit("a",256,"b",255,"c",254,
                                            tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false));
    DO_TEST(test_struct_with_array_with_bit("a",255,"b",256,"c",255,
                                            tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false));
    DO_TEST(test_struct_with_array_with_bit("a",255,"b",254,"c",255,
                                            tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false,tsf_false,tsf_true,tsf_false));
    
    DO_TEST(test_array_with_struct_with_float(0,"a","b","c",
                                              1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0));
    DO_TEST(test_array_with_struct_with_float(1,"a","b","c",
                                              1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0));
    DO_TEST(test_array_with_struct_with_float(5,"a","b","c",
                                              1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0));
    
    DO_TEST(test_struct_with_struct_with_string("d","e","f",
                                                "a",0,"b",0,"c",0,
                                                "a",0,"b",0,"c",0,
                                                "a",0,"b",0,"c",0));
    DO_TEST(test_struct_with_struct_with_string("d","e","f",
                                                "a",0,"b",1,"c",0,
                                                "a",1,"b",0,"c",1,
                                                "a",0,"b",1,"c",0));
    DO_TEST(test_struct_with_struct_with_string("d","e","f",
                                                "a",1,"b",0,"c",1,
                                                "a",0,"b",1,"c",0,
                                                "a",1,"b",0,"c",1));
    
    DO_TEST(test_choice_with_void(UINT32_MAX,NULL));
    DO_TEST(test_choice_with_void(0,"a",NULL));
    DO_TEST(test_choice_with_void(UINT32_MAX,"a",NULL));
    DO_TEST(test_choice_with_void(0,"a","",NULL));
    DO_TEST(test_choice_with_void(1,"a","",NULL));
    DO_TEST(test_choice_with_void(UINT32_MAX,"a","",NULL));
    
    DO_TEST(test_choice_with_bit("a","b","c",0,tsf_false));
    DO_TEST(test_choice_with_bit("a","b","c",1,tsf_true));
    DO_TEST(test_choice_with_bit("a","b","c",2,tsf_false));
    DO_TEST(test_choice_with_bit("a","b","c",UINT32_MAX,tsf_true));
    
    DO_TEST(test_choice_with_string("a","b","c",0,5));
    DO_TEST(test_choice_with_string("a","b","c",1,6));
    DO_TEST(test_choice_with_string("a","b","c",2,5));
    DO_TEST(test_choice_with_string("a","b","c",UINT32_MAX,6));
    
    DO_TEST(test_struct_with_choice(1,0,0));
    DO_TEST(test_struct_with_choice(1,0,UINT32_MAX));
    DO_TEST(test_struct_with_choice(2,0,0));
    DO_TEST(test_struct_with_choice(2,0,1));
    DO_TEST(test_struct_with_choice(2,0,UINT32_MAX));
    DO_TEST(test_struct_with_choice(3,0,0));
    DO_TEST(test_struct_with_choice(3,0,1));
    DO_TEST(test_struct_with_choice(3,0,2));
    DO_TEST(test_struct_with_choice(3,0,UINT32_MAX));

    DO_TEST(test_struct_with_choice(2,1,0));
    DO_TEST(test_struct_with_choice(2,1,1));
    DO_TEST(test_struct_with_choice(2,1,UINT32_MAX));
    DO_TEST(test_struct_with_choice(3,1,0));
    DO_TEST(test_struct_with_choice(3,1,1));
    DO_TEST(test_struct_with_choice(3,1,2));
    DO_TEST(test_struct_with_choice(3,1,UINT32_MAX));
    DO_TEST(test_struct_with_choice(3,2,0));
    DO_TEST(test_struct_with_choice(3,2,1));
    DO_TEST(test_struct_with_choice(3,2,2));
    DO_TEST(test_struct_with_choice(3,2,UINT32_MAX));
}

static void really_do_all_tests() {
#ifdef USE_FD_WRITER_FD_READER
    TEST_PUSH("tsf_fd_writer/tsf_fd_reader");
    reader_to_use=tsf_fd_reader;
    writer_to_use=tsf_fd_writer;
    do_all_tests();
    TEST_POP();
#endif
    
#ifdef USE_BUFFERED_WRITER_FD_READER
    TEST_PUSH("tsf_buf_write_write/tsf_fd_reader");
    reader_to_use=tsf_fd_reader;
    writer_to_use=tsf_buf_writer_write;
    do_all_tests();
    TEST_POP();
#endif
    
#ifdef USE_BUFFERED_WRITER_BUFFERED_READER
    TEST_PUSH("tsf_buf_writer_write/tsf_buf_reader_read");
    reader_to_use=tsf_buf_reader_read;
    writer_to_use=tsf_buf_writer_write;
    do_all_tests();
    TEST_POP();
#endif
    
#ifdef USE_FD_WRITER_BUFFERED_READER
    TEST_PUSH("tsf_fd_writer/tsf_buf_reader_read");
    reader_to_use=tsf_buf_reader_read;
    writer_to_use=tsf_fd_writer;
    do_all_tests();
    TEST_POP();
#endif
}

int main(int c,char **v) {
    TEST_INIT(c,v);
    
#ifdef USES_GENERATION
    if (gpc_gpint_supported()) {
        TEST_PUSH("GPINT");
        putenv(strdup("TSF_ENGINE_CONFIG=GPINT"));
        really_do_all_tests();
        TEST_POP();
    }
    
    if (gpc_dyncc_supported()) {
        TEST_PUSH("DYNCC");
        putenv(strdup("TSF_ENGINE_CONFIG=DYNCC"));
        really_do_all_tests();
        TEST_POP();
    }
#else
    really_do_all_tests();
#endif
    
    TEST_DONE();
    
    return 0;
}

