/*
 * $Header: /Users/pizlo/Repository/tsf/ppc_check_64.h,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * ppc_check_64.h -- checks if we're on a 64-bit PowerPC
 * by Filip Pizlo, 2004, 2005
 */

#ifndef FP_PPC_CHECK_64_H
#define FP_PPC_CHECK_64_H

/* tells you if we're on a 64-bit PowerPC */
static int ppc_check_64() {
    int is_32;
    
    /* note that the code below uses r3 and r4 directly instead of
       allowing the compiler to decide what registers I'm going to use.
       This is necessary, because the whole point is that I don't know
       what is the size of the registers.   Using those registers directly
       is perfectly OK, because: I know that it won't impact the ABI, and
       I know that if is_32 gets allocated to one of those registers, things
       will just work because is_32 only comes alive after my usage of r3
       and r4 end. */
    
    /* note furthermore that I'm using the long form of cmpl, which some
       assemblers may reject without being given some additional switches. */
    
#if ja_prepend_r_to_regs

    __asm__ __volatile__ ("\tli r3,-1\n"
                          "\tli r4,32767\n"
                          "\tori r4,r4,32768\n"
                          "\toris r4,r4,65535\n"
                          "\tcmpl 7,1,r3,r4\n"
                          "\tmfcr %0\n"
                          "\tandi. %0,%0,2"
                          /* outputs: */  : "=r"(is_32)
                          /* inputs: */   :
                          /* clobbers: */ : "cr7");

#else
    
    __asm__ __volatile__ ("\tli 3,-1\n"
                          "\tli 4,32767\n"
                          "\tori 4,4,32768\n"
                          "\toris 4,4,65535\n"
                          "\tcmpl 7,1,3,4\n"
                          "\tmfcr %0\n"
                          "\tandi. %0,%0,2"
                          /* outputs: */  : "=r"(is_32)
                          /* inputs: */   :
                          /* clobbers: */ : "cr7");

#endif
    
    return !is_32;
}

#endif

