/*
 * $Header: /Users/pizlo/Repository/tsf/reflect_create_test.h,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * reflect_create_test.h -- functions for creating test data using reflection
 * by Filip Pizlo, 2004, 2005
 *
 * Note: this is not a normal header file.
 */

typedef tsf_reflect_t create_data_t;

static tsf_buffer_t *create_make_buffer(tsf_type_t *type,
                                        tsf_reflect_t *r) {
    tsf_buffer_t *buf;
    TAT(buf=tsf_reflect_make_buffer(r));
    return buf;
}

static void create_destroy_data(tsf_reflect_t *r) {
    tsf_reflect_destroy(r);
}

#define DEFINE_PRIMITIVE_CREATE_FUNCS(type_lc,type_uc)              \
    static tsf_reflect_t *create_##type_lc(type_lc value) {         \
        tsf_reflect_t *r;                                           \
        TAT(r=tsf_reflect_create_##type_lc(value));                 \
        return r;                                                   \
    }                                                               \
    typedef int define_primitive_create_funcs_##type_lc

static tsf_reflect_t *create_bit(tsf_bool_t value) {
    tsf_reflect_t *r;
    TAT(r=tsf_reflect_create_bit(value));
    return r;
}

static tsf_reflect_t *create_void(void) {
    tsf_reflect_t *r;
    TAT(r=tsf_reflect_create_void());
    return r;
}

static tsf_reflect_t *create_string(uint32_t size) {
    tsf_reflect_t *r;
    unsigned char *string;
    uint32_t i;
    TAE(string=malloc(size+1));
    for (i=0;i<size;++i) {
        string[i]=(i%255)+1;
    }
    string[size]=0;
    TAT(r=tsf_reflect_create_string(string));
    TA(!strcmp(tsf_string_reflect_get(r),string));
    return r;
}

static tsf_reflect_t *create_any(tsf_buffer_t *sub) {
    tsf_reflect_t *r;
    TAT(r=tsf_reflect_create_any(sub));
    return r;
}

static tsf_reflect_t *create_array(tsf_type_t *ele_type,
                                   uint32_t total_elements,
                                   uint32_t num_elements,
                                   tsf_reflect_t **ele) {
    tsf_reflect_t *r;
    uint32_t n,i=0;
    TAT(r=tsf_reflect_create(tsf_type_create_array(ele_type)));
    for (n=total_elements;n-->0;) {
        TAT(tsf_array_reflect_append(r,tsf_reflect_clone(ele[i])));
        i=((i+1)%num_elements);
    }
    TA(tsf_array_reflect_get_num_elements(r)==total_elements);
    return r;
}

static tsf_reflect_t *create_struct(tsf_type_t *type,
                                    tsf_reflect_t **ele) {
    uint32_t i;
    tsf_reflect_t *r;
    TAT(r=tsf_reflect_create(type));
    for (i=0;i<tsf_struct_type_get_num_elements(type);++i) {
        TAT(tsf_struct_reflect_set_element_by_index(r,
                                                    i,
                                                    ele[i]));
    }
    return r;
}

static tsf_reflect_t *create_choice(tsf_type_t *type,
                                    uint32_t value,
                                    tsf_reflect_t *data) {
    tsf_reflect_t *r;
    TAT(r=tsf_reflect_create(type));
    TAT(tsf_choice_reflect_set_by_index(r,
                                        value,
                                        data));
    return r;
}



