/*
 * $Header: /Users/pizlo/Repository/tsf/reflect_verify_test.h,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * reflect_verify_test.h -- functions for verifying test data using reflection
 * by Filip Pizlo, 2004, 2005
 *
 * Note: this is not a normal header file.
 */

typedef tsf_reflect_t verify_data_t;

static tsf_reflect_t *verify_from_buffer(tsf_type_t *type,
                                         tsf_buffer_t *buf) {
    tsf_reflect_t *r;
    TAT(r=tsf_reflect_from_buffer(buf));
    return r;
}

static void verify_destroy_data(tsf_reflect_t *data) {
    tsf_reflect_destroy(data);
}

#define DEFINE_PRIMITIVE_VERIFY_FUNCS(type_lc,type_uc)              \
    static void verify_##type_lc(tsf_reflect_t *r,type_lc value) {  \
        TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))          \
           ==type_uc);                                              \
        TA(tsf_primitive_reflect_get(r,type_lc)==value);            \
    }                                                               \
    typedef int define_primitive_verify_funcs_##type_lc

static void verify_bit(tsf_reflect_t *r,tsf_bool_t value) {
    TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
       ==TSF_TK_BIT);
    TA(tsf_bit_reflect_get(r)==value);
}

static void verify_void(tsf_reflect_t *r) {
    TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
       ==TSF_TK_VOID);
}

static void verify_string(tsf_reflect_t *r,
                          uint32_t size) {
    const unsigned char *string;
    uint32_t i;
    TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
       ==TSF_TK_STRING);
    string=tsf_string_reflect_get(r);
    TA(string!=NULL);
    TA(strlen(string)==size);
    for (i=0;i<size;++i) {
        TA(string[i]==(i%255)+1);
    }
}

static tsf_buffer_t *verify_any(tsf_reflect_t *r) {
    TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
       ==TSF_TK_ANY);
    return tsf_any_reflect_get_buffer(r);
}

static void verify_array(tsf_type_t *t, /* this is the type of the array, not
                                           the type of the array element. */
                         tsf_reflect_t *r,
                         uint32_t total_elements,
                         uint32_t num_elements,
                         void (*cback)(tsf_reflect_t *r,
                                       uint32_t i,
                                       void *arg),
                         void *arg) {
    uint32_t i;
    TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
       ==TSF_TK_ARRAY);
    TA(tsf_array_reflect_get_num_elements(r)==total_elements);
    for (i=0;i<total_elements;++i) {
        cback(tsf_array_reflect_get_element(r,i),
              i%num_elements,
              arg);
    }
}

static void verify_struct(tsf_type_t *t,
                          tsf_reflect_t *r,
                          void (*cback)(tsf_reflect_t *r,
                                        uint32_t i,
                                        void *arg),
                          void *arg) {
    uint32_t i;
    TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
       ==TSF_TK_STRUCT);
    for (i=0;i<tsf_struct_type_get_num_elements(t);++i) {
        cback(tsf_struct_reflect_get_element_by_index(r,i),
              i,
              arg);
    }
}

static tsf_reflect_t *verify_choice(tsf_type_t *t,
                                    tsf_reflect_t *r,
                                    uint32_t value) {
    TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
       ==TSF_TK_CHOICE);
    TA(tsf_choice_reflect_get_index(r)
       ==value);
    if (value!=UINT32_MAX) {
        return tsf_choice_reflect_get_data(r);
    }
    return NULL;
}


