/*
 * $Header: /Users/pizlo/Repository/tsf/region_test.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * region_test.c -- tests the region allocation stuff.
 * by Filip Pizlo, 2003, 2004, 2005
 *
 * these tests are white-box.  they're also pretty lame.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/types.h>

static uint32_t malloc_cnt,
                free_cnt,
                call_cnt;

static void *my_malloc(size_t size) {
    malloc_cnt++;
    return malloc(size);
}

static void my_free(void *arg) {
    free_cnt++;
    free(arg);
}

static void my_function(void *arg) {
    call_cnt++;
}

#define malloc my_malloc
#define free my_free
#include "tsf_region.h"
#undef malloc
#undef free

/* include this after tsf_region.h is included so that tsf_region.h
 * is appropriately affected by the preprocessor hacking we are doing.
 * since test.h includes tsf.h which includes tsf_region.h, if we
 * included this up above, then the preprocessor hack would not work. */
#include "tsf_test.h"

#define SET_UP()                                \
    malloc_cnt=0;                               \
    free_cnt=0;                                 \
    call_cnt=0

#define TEAR_DOWN() if (0) printf("Impeach Bush\n")

void test_n_small_allocs(uint32_t num_allocs) {
    int *x;
    uint32_t i;
    
    SET_UP();
    
    TAE(x=tsf_region_create(sizeof(int)));
    TA(malloc_cnt==1);
    TA(free_cnt==0);
    
    *x=0;
    
    TA(*x==0);
    
    for (i=1;i<num_allocs;++i) {
        int *y;
        TAE(y=tsf_region_alloc(x,sizeof(int)));
        *y=i;
        
        TA(*x==0);
        TA(*y==(signed)i);
    }
    
    tsf_region_free(x);
    TA(malloc_cnt==1);
    TA(free_cnt==1);
    
    TEAR_DOWN();
}

void test_n_medium_allocs(uint32_t num_allocs,
                          size_t size_of_alloc,
                          uint32_t real_expected_malloc_cnt) {
    void *root;
    uint32_t i;
    size_t region_fill=size_of_alloc;
    uint32_t expected_malloc_cnt=1;
    
    SET_UP();
    
    TAE(root=tsf_region_create(size_of_alloc));

    TA(malloc_cnt==expected_malloc_cnt);
    TA(free_cnt==0);
    
    for (i=1;i<num_allocs;++i) {
        void *obj;
        TAE(obj=tsf_region_alloc(root,size_of_alloc));
        
        region_fill+=size_of_alloc;
        if (region_fill>TSF_REGION_SIZE_UNIT) {
            expected_malloc_cnt++;
            region_fill=size_of_alloc;
        }
        
        TA(malloc_cnt==expected_malloc_cnt);
        TA(free_cnt==0);
    }
    
    tsf_region_free(root);
    TA(malloc_cnt==expected_malloc_cnt);
    TA(expected_malloc_cnt==real_expected_malloc_cnt);
    TA(free_cnt==expected_malloc_cnt);
    
    TEAR_DOWN();
}

void test_n_large_allocs(uint32_t num_allocs,
                         size_t size_of_alloc) {
    void *root;
    uint32_t i;
    
    SET_UP();
    
    TAE(root=tsf_region_create(size_of_alloc));

    TA(malloc_cnt==1);
    TA(free_cnt==0);
    
    for (i=1;i<num_allocs;++i) {
        void *obj;
        TAE(obj=tsf_region_alloc(root,size_of_alloc));
        
        TA(malloc_cnt==i+1);
        TA(free_cnt==0);
    }
    
    tsf_region_free(root);
    TA(malloc_cnt==num_allocs);
    TA(free_cnt==num_allocs);
    
    TEAR_DOWN();
}



void test_callbacks(uint32_t num_cbacks) {
    void *root;
    uint32_t i;
    
    SET_UP();
    
    TAE(root=tsf_region_create(0));
    
    for (i=0;i<num_cbacks;++i) {    
        TAE(tsf_region_add_cback(root,my_function,NULL));
    }
    
    tsf_region_free(root);
    
    TA(malloc_cnt==(num_cbacks?2:1));
    TA(free_cnt==(num_cbacks?2:1));
    TA(call_cnt==num_cbacks);
    
    TEAR_DOWN();
}

#define DEFINE_TYPE_ALLOC_TESTS(type)                           \
void test_alloc_##type() {                                      \
    type *x;                                                    \
                                                                \
    SET_UP();                                                   \
                                                                \
    TAE(x=tsf_region_create(sizeof(type)));                     \
                                                                \
    *x=5;                                                       \
                                                                \
    TA(*x==5);                                                  \
                                                                \
    tsf_region_free(x);                                         \
                                                                \
    TEAR_DOWN();                                                \
}                                                               \
                                                                \
void test_alloc_##type##_after_char() {                         \
    char *c;                                                    \
    type *x;                                                    \
                                                                \
    SET_UP();                                                   \
                                                                \
    TAE(c=tsf_region_create(sizeof(char)));                     \
    TAE(x=tsf_region_alloc(c,sizeof(type)));                    \
                                                                \
    *x=5;                                                       \
    TA(*x==5);                                                  \
                                                                \
    tsf_region_free(c);                                         \
    TEAR_DOWN();                                                \
}

DEFINE_TYPE_ALLOC_TESTS(int8_t)
DEFINE_TYPE_ALLOC_TESTS(uint8_t)
DEFINE_TYPE_ALLOC_TESTS(int16_t)
DEFINE_TYPE_ALLOC_TESTS(uint16_t)
DEFINE_TYPE_ALLOC_TESTS(int32_t)
DEFINE_TYPE_ALLOC_TESTS(uint32_t)
DEFINE_TYPE_ALLOC_TESTS(int64_t)
DEFINE_TYPE_ALLOC_TESTS(uint64_t)
DEFINE_TYPE_ALLOC_TESTS(float)
DEFINE_TYPE_ALLOC_TESTS(double)

#define N_REPEAT 10

int main(int c,char **v) {
    TEST_INIT(c,v);
    
    /* really simple tests */
    DO_TEST(REPEAT(test_n_small_allocs(1),N_REPEAT));
    
    /* actually do multiple allocs */
    DO_TEST(REPEAT(test_n_small_allocs(2),N_REPEAT));
    DO_TEST(REPEAT(test_n_small_allocs(10),N_REPEAT));
    
    /* tests that require multiple malloc calls */
    DO_TEST(REPEAT(test_n_medium_allocs(1,TSF_REGION_SIZE_UNIT/2-1,1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(2,TSF_REGION_SIZE_UNIT/2-1,1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(3,TSF_REGION_SIZE_UNIT/2-1,2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(10,TSF_REGION_SIZE_UNIT/2-1,5),
                   N_REPEAT));
    
    DO_TEST(REPEAT(test_n_medium_allocs(1,TSF_REGION_SIZE_UNIT/2,1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(2,TSF_REGION_SIZE_UNIT/2,1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(3,TSF_REGION_SIZE_UNIT/2,2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(10,TSF_REGION_SIZE_UNIT/2,5),
                   N_REPEAT));
    
    DO_TEST(REPEAT(test_n_medium_allocs(1,TSF_REGION_SIZE_UNIT/2+1,1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(2,TSF_REGION_SIZE_UNIT/2+1,2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(3,TSF_REGION_SIZE_UNIT/2+1,3),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(10,TSF_REGION_SIZE_UNIT/2+1,10),
                   N_REPEAT));
    
    DO_TEST(REPEAT(test_n_medium_allocs(1,TSF_REGION_SIZE_UNIT-1,1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(2,TSF_REGION_SIZE_UNIT-1,2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(3,TSF_REGION_SIZE_UNIT-1,3),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(10,TSF_REGION_SIZE_UNIT-1,10),
                   N_REPEAT));
    
    DO_TEST(REPEAT(test_n_medium_allocs(1,TSF_REGION_SIZE_UNIT,1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(2,TSF_REGION_SIZE_UNIT,2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(3,TSF_REGION_SIZE_UNIT,3),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_medium_allocs(10,TSF_REGION_SIZE_UNIT,10),
                   N_REPEAT));
    
    /* finally, test large allocs */
    DO_TEST(REPEAT(test_n_large_allocs(1,TSF_REGION_SIZE_UNIT+1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(2,TSF_REGION_SIZE_UNIT+1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(3,TSF_REGION_SIZE_UNIT+1),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(10,TSF_REGION_SIZE_UNIT+1),
                   N_REPEAT));
    
    DO_TEST(REPEAT(test_n_large_allocs(1,TSF_REGION_SIZE_UNIT*2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(2,TSF_REGION_SIZE_UNIT*2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(3,TSF_REGION_SIZE_UNIT*2),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(10,TSF_REGION_SIZE_UNIT*2),
                   N_REPEAT));
    
    DO_TEST(REPEAT(test_n_large_allocs(1,TSF_REGION_SIZE_UNIT*10),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(2,TSF_REGION_SIZE_UNIT*10),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(3,TSF_REGION_SIZE_UNIT*10),
                   N_REPEAT));
    DO_TEST(REPEAT(test_n_large_allocs(10,TSF_REGION_SIZE_UNIT*10),
                   N_REPEAT));
    
    /* type tests; these help catch alignment errors */
    DO_TEST(test_alloc_int8_t());
    DO_TEST(test_alloc_uint8_t());
    DO_TEST(test_alloc_int16_t());
    DO_TEST(test_alloc_uint16_t());
    DO_TEST(test_alloc_int32_t());
    DO_TEST(test_alloc_uint32_t());
    DO_TEST(test_alloc_int64_t());
    DO_TEST(test_alloc_uint64_t());
    DO_TEST(test_alloc_float());
    DO_TEST(test_alloc_double());
    
    DO_TEST(test_alloc_int8_t_after_char());
    DO_TEST(test_alloc_uint8_t_after_char());
    DO_TEST(test_alloc_int16_t_after_char());
    DO_TEST(test_alloc_uint16_t_after_char());
    DO_TEST(test_alloc_int32_t_after_char());
    DO_TEST(test_alloc_uint32_t_after_char());
    DO_TEST(test_alloc_int64_t_after_char());
    DO_TEST(test_alloc_uint64_t_after_char());
    DO_TEST(test_alloc_float_after_char());
    DO_TEST(test_alloc_double_after_char());
    
    /* callback tests. */
    DO_TEST(test_callbacks(0)); /* sanity check */
    DO_TEST(test_callbacks(1));
    DO_TEST(test_callbacks(2));
    DO_TEST(test_callbacks(20));
    
    TEST_DONE();
    
    return 0;
}



