/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_buf_writer.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_buf_writer.c -- implementation of buffered writer
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "tsf_format.h"

#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>

tsf_buf_wtr_t *tsf_buf_writer_create(int fd,
                                     uint32_t buf_size,
                                     tsf_bool_t keep_fd) {
    tsf_buf_wtr_t *result;
    
    result=malloc(sizeof(tsf_buf_wtr_t));
    if (result==NULL) {
        tsf_set_errno("Could not malloc tsf_buf_wtr_t");
        return NULL;
    }
    
    result->buf=malloc(buf_size);
    if (result->buf==NULL) {
        free(result);
        tsf_set_errno("Could not malloc buffer");
        return NULL;
    }

    result->fd=fd;
    result->keep_fd=keep_fd;
    result->size=buf_size;
    result->pos=0;
    
    return result;
}

void tsf_buf_writer_destroy(tsf_buf_wtr_t *buf) {
    free(buf->buf);
    free(buf);
}

tsf_bool_t tsf_buf_writer_write(void *_buf,
                                void *data,
                                uint32_t len) {
#ifdef HAVE_WRITEV
    tsf_buf_wtr_t *buf=_buf;
    
    if (buf->pos+len>buf->size) {
        if (buf->pos) {
            struct iovec iov[2];
            iov[0].iov_base=(void*)buf->buf;
            iov[0].iov_len=buf->pos;
            iov[1].iov_base=data;
            iov[1].iov_len=len;
            for (;;) {
                int res=writev(buf->fd,iov,2);
                if (res<0) {
                    if (errno==EINTR) {
                        continue;
                    }
                    tsf_set_errno("Calling writev() in "
                                  "tsf_buf_writer_write()");
                    return tsf_false;
                }
                if ((unsigned)res>=iov[0].iov_len) {
                    data=((uint8_t*)data)+res-iov[0].iov_len;
                    len-=res-iov[0].iov_len;
                    break;
                }
                iov[0].iov_base+=res;
                iov[0].iov_len-=res;
            }
            buf->pos=0;
            if (len==0) {
                return tsf_true;
            }
        }
        return tsf_fd_writer(&buf->fd,data,len);
    }
#else
    if (buf->pos+len>=buf->size) {
        /* have to flush because there isn't enough room in our buffer
         * otherwise. */
        if (!tsf_buf_writer_flush(buf)) {
            return tsf_false;
        }
    }
    
    if (len>=buf->size&&buf->pos==0) {
        /* we just write now because we are being given more data than
         * our buffer could store. */
        return tsf_fd_writer(&buf->fd,data,len);
    }
#endif
    
    /*
    printf(" buf->buf = %p\n",buf->buf);
    printf(" buf->pos = " fui32 "\n",buf->pos);
    printf("buf->size = " fui32 "\n",buf->size);
    printf("     data = %p\n",data);
    printf("      len = " fui32 "\n",len);
    */
    
    memcpy(buf->buf+buf->pos,data,len);
    buf->pos+=len;
    
    return tsf_true;
}

tsf_bool_t tsf_buf_writer_flush(tsf_buf_wtr_t *buf) {
    if (!tsf_fd_writer(&buf->fd,buf->buf,buf->pos)) {
        return tsf_false;
    }
    buf->pos=0;
    return tsf_true;
}



