/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_config.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_config.c -- simple program that spits out configuration options for
 *                 the compiler and linker.
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal_config.h"
#include "tsf_build_defines.h" /* not a real header file; gives us DYNCC_BUILD_CMD_SUFFIX */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static const char *usage_msg=
"Usage: tsf_config <commands> ...\n"
"Where the available commands are:\n"
"  --cppflags     Lists the flags you would want to pass to a C preprocessor\n"
"                 when compiling a program that uses TSF.\n"
"  --cflags       Lists the flags you would want to pass to a C compiler when\n"
"                 compiling a program that uses TSF.\n"
"  --cxxflags     Lists the flags you would want to pass to a C++ compiler when\n"
"                 compiling a program that uses TSF.\n"
"  --ldflags      Lists the flags you would want to pass to the linker when\n"
"                 linking a program that uses TSF.\n"
"  --libs         Lists the library link commands that you would want to pass to\n"
"                 the linker when linking a program that uses TSF.\n"
"  --prefix       Prints the TSF install prefix.\n"
"  --version      Prints the TSF version that this program corresponds to.\n";

void usage() {
    fprintf(stderr,"%s",usage_msg);
    exit(1);
}

int main(int c,char **v) {
    int i;
    int print_usage=0;
    
    for (i=1;i<c;++i) {
        if (!strcmp(v[i],"--cppflags")) {
            printf("-I%s %s %s\n",INCLUDE_DIR,XTRA_CPPFLAGS,CPPFLAGS);
        } else if (!strcmp(v[i],"--cflags")) {
            printf("%s\n",CFLAGS);
        } else if (!strcmp(v[i],"--cxxflags")) {
            printf("%s\n",CFLAGS);
        } else if (!strcmp(v[i],"--ldflags")) {
            printf("-L%s %s %s\n",LIB_DIR,XTRA_LDFLAGS,LDFLAGS);
        } else if (!strcmp(v[i],"--libs")) {
            printf("-ltsf %s %s\n",XTRA_LIBS,LIBS);
        } else if (!strcmp(v[i],"--prefix")) {
            printf("%s\n",PREFIX);
        } else if (!strcmp(v[i],"--version")) {
            printf("%s\n",PACKAGE_STRING);
        } else {
            fprintf(stderr,"%s is not an option.\n",v[i]);
            print_usage=1;
        }
    }
    
    if (print_usage) {
        fprintf(stderr,"\n");
    }
    
    if (c<2 || print_usage) {
        usage();
    }
    
    return 0;
}


