/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_copier.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_copier.c -- implementation of the copier thingy
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "gpc.h"

tsf_copier_t *tsf_copier_create(tsf_type_t *dest,
                                tsf_type_t *src,
                                tsf_engine_t conf,
				const char *cache_dir) {
    tsf_copier_t *result;
    gpc_proto_t *proto=
        tsf_gpc_code_gen_generate_converter(dest,src);
    
    if (proto==NULL) {
        return NULL;
    }
    
    gpc_code_gen_debug("converter",proto);
    
    result=malloc(sizeof(tsf_copier_t));
    if (result==NULL) {
        tsf_set_errno("Could not malloc tsf_copier_t");
        gpc_proto_destroy(proto);
        return NULL;
    }
    
    result->copier=
	gpc_program_from_proto(proto,conf,cache_dir);
    
    gpc_proto_destroy(proto);
    
    if (result->copier==NULL) {
        free(result);
        return NULL;
    }
    
    return result;
}

void tsf_copier_destroy(tsf_copier_t *copier) {
    gpc_program_destroy(copier->copier);
    free(copier);
}

void *tsf_copier_clone(tsf_copier_t *copier,void *src) {
    gpc_cell_t args[2];
    args[0]=(gpc_cell_t)NULL;
    args[1]=(gpc_cell_t)src;
    return (void*)gpc_program_run(copier->copier,NULL,args);
}

void *tsf_copier_clone_in(tsf_copier_t *copier,void *src,void *region) {
    gpc_cell_t args[2];
    args[0]=(gpc_cell_t)NULL;
    args[1]=(gpc_cell_t)src;
    return (void*)gpc_program_run(copier->copier,region,args);
}

tsf_bool_t tsf_copier_copy(tsf_copier_t *copier,
                           void *dest,
                           void *src,
                           void *region) {
    gpc_cell_t args[2];
    args[0]=(gpc_cell_t)dest;
    args[1]=(gpc_cell_t)src;
    return gpc_program_run(copier->copier,region,args)!=0?
               tsf_true:tsf_false;
}




