/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_generator.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_generator.c -- the thing that takes native data and generates
 *                    a buffer.
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "gpc.h"
#include <unistd.h>
#include <stdio.h>
#include <string.h>

tsf_genrtr_t *tsf_generator_create(tsf_type_t *type,
                                   tsf_engine_t conf,
				   const char *cache_dir) {
    tsf_genrtr_t *result;
    gpc_proto_t *proto;
    
    proto=tsf_gpc_code_gen_generate_generator(type);
    if (proto==NULL) {
        return NULL;
    }
    
    gpc_code_gen_debug("generator",proto);
    
    result=malloc(sizeof(tsf_genrtr_t));
    if (result==NULL) {
        tsf_set_errno("Could not malloc tsf_genrtr_t");
        gpc_proto_destroy(proto);
        return NULL;
    }
    
    result->type=tsf_type_dup(type);
    result->generator=
	gpc_program_from_proto(proto,conf,cache_dir);
    
    gpc_proto_destroy(proto);
    
    if (result->generator==NULL) {
        free(result);
        return NULL;
    }
    
    return result;
}

void tsf_generator_destroy(tsf_genrtr_t *gen) {
    gpc_program_destroy(gen->generator);
    tsf_type_destroy(gen->type);
    free(gen);
}

tsf_buffer_t *tsf_generator_generate(tsf_genrtr_t *gen,
                                     void *data) {
    gpc_cell_t args[2];
    args[0]=(gpc_cell_t)gen->type;
    args[1]=(gpc_cell_t)data;
    return (tsf_buffer_t*)gpc_program_run(gen->generator,NULL,args);
}

tsf_buffer_t *tsf_generator_generate_in(tsf_genrtr_t *gen,
                                        void *data,
                                        void *region) {
    gpc_cell_t args[2];
    args[0]=(gpc_cell_t)gen->type;
    args[1]=(gpc_cell_t)data;
    return (tsf_buffer_t*)gpc_program_run(gen->generator,region,args);
}


