/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_indent.c,v 1.2 2007/07/08 08:15:01 pizlo Exp $
 * 
 * tsf_indent.c -- manages indentation for code generation
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"

/* number of spaces per 'tab' */
#define SPACES_PER_TAB 4

/* number of spaces plus one for zero character */
#define SPACES_SIZE 81

static char spaces[SPACES_SIZE];

static void real_init_indent() {
    uint32_t i;
    for (i=0;i<SPACES_SIZE-1;++i) {
        spaces[i]=' ';
    }
    spaces[SPACES_SIZE-1]=0;
}

#if TSF_HAVE_PTHREAD
#include <pthread.h>

static pthread_once_t once_control = PTHREAD_ONCE_INIT;

static void init_indent() {
    /* NOTE: pthread_once can be slow on some platforms.  But it shouldn't
       matter here, since indentation is not that big of a deal (the
       library itself doesn't use it). */
    pthread_once(&once_control,real_init_indent);
}
#else
tsf_bool_t spaces_inited=tsf_false;

static void init_indent() {
    if (!spaces_inited) {
        real_init_indent();
        spaces_inited=tsf_true;
    }
}
#endif

const char *tsf_indent_str(uint32_t indent) {
    init_indent();
    if (indent*SPACES_PER_TAB>SPACES_SIZE) {
        indent=SPACES_SIZE/SPACES_PER_TAB;
    }
    
    return spaces+SPACES_SIZE-1-indent*SPACES_PER_TAB;
}

