/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_named_type.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_named_type.c -- stores named types
 * by Filip Pizlo, 2004, 2005
 */

#include "tsf_internal.h"

tsf_named_type_t *tsf_named_type_create(const char *name,
                                        tsf_type_t *type,
                                        uint32_t index) {
    tsf_named_type_t *node=malloc(sizeof(tsf_named_type_t));
    if (node==NULL) {
        tsf_set_errno("Could not malloc tsf_named_type_t");
        goto failure_1;
    }
    
    if (strlen(name)+1>TSF_STR_SIZE) {
        tsf_set_error(TSF_E_STR_OVERFLOW,
                      "String '%s' for too long to be used as an element "
                      "name",name);
        goto failure_2;
    }
    
    node->name=strdup(name);
    if (node->name==NULL) {
        tsf_set_errno("Could not allocate name");
        goto failure_2;
    }
    
    node->type=type;
    node->optional=tsf_false;
    node->has_offset=tsf_false;
    node->offset=0;
    node->index=index;
    node->comment=NULL;
    
    return node;
    
failure_2:
    free(node);
failure_1:
    tsf_type_destroy(type);
    return NULL;
}

void tsf_named_type_destroy(tsf_named_type_t *node) {
    tsf_type_destroy(node->type);
    free(node->name);
    if (node->comment!=NULL) {
        free(node->comment);
    }
    free(node);
}

tsf_bool_t tsf_named_type_set_comment(tsf_named_type_t *node,
                                      const char *comment) {
    char *comment_copy;
    
    if (strlen(comment)+1>TSF_MAX_COMMENT_SIZE) {
        tsf_set_error(TSF_E_STR_OVERFLOW,
                      "Comment string too long.");
        return tsf_false;
    }
    
    comment_copy=strdup(comment);
    if (comment_copy==NULL) {
        tsf_set_errno("Could not strcpy comment");
        return tsf_false;
    }
    free(node->comment);
    node->comment=comment_copy;
    
    return tsf_true;
}

int tsf_named_type_get_hash(tsf_named_type_t *node) {
    return tsf_st_strhash(node->name)|
           (tsf_type_get_hash(node->type)<<8)|
	   node->optional<<16;
}



