/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_safety.c,v 1.2 2007/07/08 08:15:01 pizlo Exp $
 *
 * tsf_safety.c -- the safety API allows you to specify when you are in code
 *                 that should not be considered for fault injection
 * by Filip Pizlo, 2004, 2005
 */

#include "tsf_internal.h"

#ifdef HAVE_PTHREAD

static pthread_once_t once_control = PTHREAD_ONCE_INIT;
static pthread_key_t in_safe_code;

static void init(void) {
    pthread_key_create(&in_safe_code,NULL);
}

tsf_bool_t tsf_set_in_safe_code(tsf_bool_t value) {
    void *result;
    
    /* NOTE: pthread_once can be slow on some platforms.  But here it
       shouldn't matter. */
    pthread_once(&once_control, init);
    
    result=pthread_getspecific(in_safe_code);
    pthread_setspecific(in_safe_code,value?&value:NULL);
    return result!=NULL;
}

tsf_bool_t tsf_in_safe_code(void) {
    pthread_once(&once_control, init);
    return pthread_getspecific(in_safe_code)!=NULL;
}

#else

static tsf_bool_t in_safe_code=tsf_true;

tsf_bool_t tsf_set_in_safe_code(tsf_bool_t value) {
    tsf_bool_t result=in_safe_code;
    in_safe_code=value;
    return result;
}

tsf_bool_t tsf_in_safe_code(void) {
    return in_safe_code;
}

#endif

