/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_type_out_map.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_type_out_map.c -- type map implementation
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"

tsf_type_out_map_t *tsf_type_out_map_create() {
    tsf_type_out_map_t *ret=tsf_st_init_typetable();
    if (ret==NULL) {
        tsf_set_errno("Could not tsf_st_init_typetable()");
        return NULL;
    }
    
    return ret;
}

static int table_destroy_pair(tsf_type_t *type,
                              uint32_t *value,
                              void *arg) {
    tsf_type_destroy(type);
    free(value);
    return TSF_ST_CONTINUE;
}

void tsf_type_out_map_destroy(tsf_type_out_map_t *type_map) {
    tsf_st_foreach(type_map,(tsf_st_func_t)table_destroy_pair,NULL);
    tsf_st_free_table(type_map);
}

tsf_bool_t tsf_type_out_map_find_type_code(tsf_type_out_map_t *type_map,
                                           tsf_type_t *type,
                                           uint32_t *type_code) {
    uint32_t *tc_ptr;
    if (!tsf_st_lookup(type_map,(char*)type,(void**)&tc_ptr)) {
        tsf_set_error(TSF_E_INTERNAL,
                      "Given type was not found in type map even though "
                      "it was asserted that the type would be found");
        return tsf_false;
    }
    if (type_code!=NULL) {
        *type_code=*tc_ptr;
    }
    return tsf_true;
}

tsf_bool_t tsf_type_out_map_get_type_code(tsf_type_out_map_t *type_map,
                                          tsf_type_t *type,
                                          uint32_t *type_code,
                                          tsf_bool_t *created) {
    uint32_t *tc_ptr;
    
    if (tsf_st_lookup(type_map,(char*)type,(void**)&tc_ptr)) {
        if (type_code!=NULL) {
            *type_code=*tc_ptr;
        }
        if (created!=NULL) {
            *created=tsf_false;
        }
    } else {
        tc_ptr=malloc(sizeof(uint32_t));
        if (tc_ptr==NULL) {
            tsf_set_errno("Could not malloc type code");
            return tsf_false;
        }
        *tc_ptr=type_map->num_entries;
        if (tsf_st_add_direct(type_map,
                              (char*)tsf_type_dup(type),
                              tc_ptr)<0) {
            tsf_set_errno("Could not tsf_st_add_direct()");
            free(tc_ptr);
            return tsf_false;
        }
        if (type_code!=NULL) {
            *type_code=*tc_ptr;
        }
        if (created!=NULL) {
            *created=tsf_true;
        }
    }
    
    return tsf_true;
}







