/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_types.h,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_types.h -- brings in all of the types we'll need
 * by Filip Pizlo, 2003, 2004, 2005
 */

#ifndef FP_TSF_TYPES_H
#define FP_TSF_TYPES_H

#include "tsf_config.h"

/* need to use this hack because NetBSD sucks. */
#include "tsf_inttypes.h"

/* our own definition of UINT32_MAX, which we need to use in all external
 * header files because C++ is dumb. */
#define TSF_UINT32_MAX ((((uint32_t)1)<<31)+\
                        ((((uint32_t)1)<<31)-1))

/* something that inttypes doesn't seem to define */
#if TSF_SIZEOF_VOID_P == 2
#define TSF_UINTPTR_C(arg) UINT16_C(arg)
#else
#if TSF_SIZEOF_VOID_P == 4
#define TSF_UINTPTR_C(arg) UINT32_C(arg)
#else
#if TSF_SIZEOF_VOID_P == 8
#define TSF_UINTPTR_C(arg) UINT64_C(arg)
#else
#error "Don't know how to express void * literals"
#endif
#endif
#endif

/* every system hacker needs sys/types! */
#include <sys/types.h>

/* our own offset type.  meant to be entirely compatible with the
 * system's offset type. */
#if TSF_SIZEOF_OFF_T == 2   /* who knows? */
typedef int16_t tsf_off_t;
#else
#if TSF_SIZEOF_OFF_T == 4
typedef int32_t tsf_off_t;
#else
#if TSF_SIZEOF_OFF_T == 8
typedef int64_t tsf_off_t;
#else
#error "Don't know how to represent file offsets"
#endif
#endif
#endif

/* why not use std_bool.h?  becuase it will fuck us if this header is
 * used from C++!  there seems to be no good way of making sure that
 * the C99 _Bool type and the C++ bool type are the same size, and,
 * what is more unnerving is that there is no good way, if one is
 * writing ANSI C to conjure up a type that is magically the same size
 * as the C++ bool type.  So, the only solution is to not use _Bool
 * or bool, and just roll our own thing.  that way, you know what size
 * it is from either C or C++. */
typedef uint8_t tsf_bool_t;

/* use this for va_arg */
typedef unsigned tsf_arg_bool_t;

/* you don't actually have to use these macros.  you can use 1
 * and 0 directly.  or if you're in C++, you should just be able
 * to use C++'s true and false keywords. */
#define tsf_true 1
#define tsf_false 0

/* a tristate type, which might come in handy */
typedef enum {
    tsf_no,
    tsf_maybe,
    tsf_yes
} tsf_tristate_t;

/* what follows are some types that must be pre-declared. */

struct gpc_proto;
typedef struct gpc_proto gpc_proto_t;

struct gpc_program;
typedef struct gpc_program gpc_program_t;

/* enumeration of supported engine types.  the first four things
 * in the enumerations are what you supply.  the last one is used
 * internally.  note that any object that has a field of type
 * tsf_engine_t is guaranteed to never hold the values
 * TSF_EN_DEFAULT or TSF_EN_NOT_SUPPORTED, although either of those
 * values may be passed to that object at one time or another. */
enum tsf_engine {
    TSF_EN_DEFAULT,
    TSF_EN_GPINT,
    TSF_EN_DYNCC,
    TSF_EN_NOT_SUPPORTED
};

/* special values for the cache directory.  note that NULL is not
   a valid value for the cache directory parameter. */
extern char const * const TSF_CD_NO_CACHE;
extern char const * const TSF_CD_DEFAULT;

typedef enum tsf_engine tsf_engine_t;

enum tsf_zip_mode {
    TSF_ZIP_NONE,
    TSF_ZIP_ZLIB,
    TSF_ZIP_BZIP2,
    TSF_ZIP_UNKNOWN
};

typedef enum tsf_zip_mode tsf_zip_mode_t;

#endif

